#include "snes9xw.h"
#include "globl.h"
#include "directx.h"
#include "joywin.h"
#include <math.h>

struct SJoypad Joypad[5] = {
	{ true,								/* Joypad 1 default keys:*/
	VK_LEFT, 0, 0, VK_RIGHT, 0, 0, VK_UP, VK_DOWN,	/* Left, Left_Up, Left_Down, Right, Right_Up, Right_Down, Up, Down */
	VK_RETURN, VK_SPACE,				/* Start, Select */
	'D', 'C',							/* A B */
	'S', 'X',							/* X Y */
	'A', 'Z',							/* L R */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	 { true,							/* Joypad 2 default joystick 0 */
	0x8000, 0, 0, 0x8001, 0, 0, 0x8002, 0x8003,		/* Left, Left_Up, Left_Down, Right, Right_Up, Right_Down, Up, Down */
	0x8008, 0x8009,						/* Start, Select */
	0x800A, 0x800B,						/* A B */
	0x800C, 0x800D,						/* X Y */
	0x800E, 0x800F,						/* L R */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	 { false,						/* Joypad 3 default off */
	0, 0, 0, 0, 0, 0, 0, 0,							/* Left, Left_Up, Left_Down, Right, Right_Up, Right_Down, Up, Down */
	0, 0,								/* Start, Select */
	0, 0,								/* A B */
	0, 0,								/* X Y */
	0, 0,								/* L R */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	 { false,						/* Joypad 4 default off */
	0, 0, 0, 0, 0, 0, 0, 0,							/* Left, Left_Up, Left_Down, Right, Right_Up, Right_Down, Up, Down */
	0, 0,								/* Start, Select */
	0, 0,								/* A B */
	0, 0,								/* X Y */
	0, 0,								/* L R */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	 { false,						/* Joypad 5 default off */
	0, 0, 0, 0, 0, 0, 0, 0,							/* Left, Left_Up, Left_Down, Right, Right_Up, Right_Down, Up, Down */
	0, 0,								/* Start, Select */
	0, 0,								/* A B */
	0, 0,								/* X Y */
	0, 0,								/* L R */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
	};

/* HIWORD == MF_xxxx Function number, LOWORD == button code */
DWORD MapFunc[13] = {MF_TURBOPAUSEFWD << 16, MF_TURBOPAUSEFWD << 16, MF_XPP01 << 16, MF_XPP02 << 16,
                    MF_XPP03 << 16, MF_XPP04 << 16, MF_XPP05 << 16, MF_XPP06 << 16, MF_XPP07 << 16,
                    MF_XPP08 << 16, MF_XPP09 << 16, MF_SSTURBO << 16, MF_SSPAUSE << 16};
static unsigned char MapFuncPress;
static unsigned char MapFuncPressed;
static unsigned char TurbofwdPressed[4];

static struct mmJoystick_private mmJoystick;
static struct diJoystick_private diJoystick;
//static struct diKeyboard_private diKeyboard;
struct diMouse_private diMouse;
static unsigned long joypads [5];

#define OSD_ANALOGMAX       ( 127.0)
#define OSD_ANALOGMIN       (-128.0)
#ifdef PI
#undef PI
#endif
#define PI 3.1415926534

static DWORD Joystick_DeadZoneMin(DWORD dwMin, DWORD dwMax, UINT nDeadZone)
{
    return (DWORD)((dwMax - dwMin) * ((100.0 - nDeadZone) / 200.0));
}

static DWORD Joystick_DeadZoneMax(DWORD dwMin, DWORD dwMax, UINT nDeadZone)
{
    return (DWORD)((dwMax - dwMin) * ((100.0 + nDeadZone) / 200.0));
}

#define SetDeadZone(Joy, Axis)                                                       \
            Joy.m_dw##Axis##DZmin = Joystick_DeadZoneMin(Joy.m_JoyCaps.w##Axis##min, \
                                                Joy.m_JoyCaps.w##Axis##max,          \
                                                Joy.m_nDeadZone);                    \
            Joy.m_dw##Axis##DZmax = Joystick_DeadZoneMax(Joy.m_JoyCaps.w##Axis##min, \
                                                Joy.m_JoyCaps.w##Axis##max,          \
                                                Joy.m_nDeadZone);

static void mmJoystick_poll_joysticks(void)
{
	int i = NUM_JOYSTICKS;

	if (mmJoystick.m_nNumJoysticks == 0)
		return;

	while (i--)
	{
		if (mmJoystick.m_Joy[i].m_bUseJoystick == TRUE)
			joyGetPosEx(mmJoystick.m_Joy[i].m_uJoyID, &mmJoystick.m_Joy[i].m_JoyInfo);
	}
}

static int mmJoystick_init()
{
	int         i;
	MMRESULT    mmResult;

	memset(&mmJoystick, 0, sizeof(struct mmJoystick_private));

	for (i = 0; i < NUM_JOYSTICKS; i++)
	{
		mmJoystick.m_Joy[i].m_bUseJoystick = TRUE;
		mmJoystick.m_Joy[i].m_nDeadZone    = 75;    /* No way to set dead zone right now */
		mmJoystick.m_Joy[i].m_uJoyID       = (JOYSTICKID1 + i);
	}

	/* User turned off joy option or a joy driver is not installed. */
	if (joyGetNumDevs() == 0) {
		for(i = 0; i < NUM_JOYSTICKS; i++)
			mmJoystick.m_Joy[i].m_bUseJoystick = FALSE;
	}

	for (i = 0; i < NUM_JOYSTICKS; i++)
	{
		if (mmJoystick.m_Joy[i].m_bUseJoystick == FALSE)
			continue;

		/* Determine if JOYSTICKID[1-16] is plugged in. */
		mmJoystick.m_Joy[i].m_JoyInfo.dwSize  = sizeof(JOYINFOEX);
		mmJoystick.m_Joy[i].m_JoyInfo.dwFlags = JOY_RETURNBUTTONS | JOY_RETURNX | JOY_RETURNY;

		mmResult = joyGetPosEx(mmJoystick.m_Joy[i].m_uJoyID, &mmJoystick.m_Joy[i].m_JoyInfo);

		if (mmResult == JOYERR_NOERROR)
		{
			joyGetDevCaps(mmJoystick.m_Joy[i].m_uJoyID, &mmJoystick.m_Joy[i].m_JoyCaps, sizeof(JOYCAPS));
			SetDeadZone(mmJoystick.m_Joy[i], X);
			SetDeadZone(mmJoystick.m_Joy[i], Y);

			/* Currently only analog support for X and Y axis. */
			mmJoystick.m_Joy[i].m_dXScale = ((OSD_ANALOGMAX - OSD_ANALOGMIN) / (mmJoystick.m_Joy[i].m_JoyCaps.wXmax - mmJoystick.m_Joy[i].m_JoyCaps.wXmin));
			mmJoystick.m_Joy[i].m_dYScale = ((OSD_ANALOGMAX - OSD_ANALOGMIN) / (mmJoystick.m_Joy[i].m_JoyCaps.wYmax - mmJoystick.m_Joy[i].m_JoyCaps.wYmin));

			/* Check for other axis */
			if (mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_HASZ)
			{
				mmJoystick.m_Joy[i].m_JoyInfo.dwFlags |= JOY_RETURNZ;
				SetDeadZone(mmJoystick.m_Joy[i], Z);
			}

			if (mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_HASR)
			{
				mmJoystick.m_Joy[i].m_JoyInfo.dwFlags |= JOY_RETURNR;
				SetDeadZone(mmJoystick.m_Joy[i], R);
			}

			if (mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_HASU)
			{
				mmJoystick.m_Joy[i].m_JoyInfo.dwFlags |= JOY_RETURNU;
				SetDeadZone(mmJoystick.m_Joy[i], U);
			}

			if (mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_HASV)
			{
				mmJoystick.m_Joy[i].m_JoyInfo.dwFlags |= JOY_RETURNV;
				SetDeadZone(mmJoystick.m_Joy[i], V);
			}

			if (mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_HASPOV
			&&  mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_POV4DIR)
			{
				mmJoystick.m_Joy[i].m_JoyInfo.dwFlags |= JOY_RETURNPOV;
			}

			mmJoystick.m_nNumJoysticks++;
		}
		else
		if (mmResult == JOYERR_UNPLUGGED)
		{
			mmJoystick.m_Joy[i].m_bUseJoystick = FALSE;
		}
		else
		{
			/* Some other error with the joystick. Don't use it. */
			mmJoystick.m_Joy[i].m_bUseJoystick = FALSE;
		}
	}

	/* Initialize JOYINFOEX */      
	mmJoystick_poll_joysticks();

	return 0;
}

static void S9xWinScanmmJoypads ();
static void S9xWinScandiJoypads ();
static void (*S9xWinScanJoypads)() = S9xWinScanmmJoypads;
static unsigned char DirectInput_init();

static WORD GetdiJoyButtonCode();
static WORD GetmmJoyButtonCode();
WORD (*GetJoyButtonCode)() = GetmmJoyButtonCode;

void SwitchInputProc()
{
	if(DirectX.DIuse) {
		S9xWinScanJoypads = S9xWinScandiJoypads;
		GetJoyButtonCode = GetdiJoyButtonCode;
	}
	else {
		S9xWinScanJoypads = S9xWinScanmmJoypads;
		GetJoyButtonCode = GetmmJoyButtonCode;
	}
}

void S9xwJoystickInit()
{
	DirectInput_exit();
	if(DirectX.DIuse) {
		if(!DirectInput_init()) {
			if(GUI.hWnd)
				MessageBox(GUI.hWnd, Languages[ GUI.Language].errInitDI, "Snes9X - DirectInput", MB_OK | MB_ICONINFORMATION);
			DirectX.DIuse = FALSE;
			DirectInput_exit();
			mmJoystick_init();
			SwitchInputProc();
			return;
		}
		SwitchInputProc();
	}
	else {
		mmJoystick_init();
		SwitchInputProc();
	}
}

static bool S9xGetStatemm (WORD KeyIdent)
{
	if (KeyIdent & 0x8000)
	{
		int i = (KeyIdent >> 8) & 15;

		switch (KeyIdent & 0xff)
		{
			case 0: return !mmJoystick.m_Joy[i].JoyState.Left;
			case 1: return !mmJoystick.m_Joy[i].JoyState.Right;
			case 2: return !mmJoystick.m_Joy[i].JoyState.Up;
			case 3: return !mmJoystick.m_Joy[i].JoyState.Down;
			case 4: return !mmJoystick.m_Joy[i].JoyState.PovLeft;
			case 5: return !mmJoystick.m_Joy[i].JoyState.PovRight;
			case 6: return !mmJoystick.m_Joy[i].JoyState.PovUp;
			case 7: return !mmJoystick.m_Joy[i].JoyState.PovDown;
			case 41:return !mmJoystick.m_Joy[i].JoyState.ZUp;
			case 42:return !mmJoystick.m_Joy[i].JoyState.ZDown;
			case 43:return !mmJoystick.m_Joy[i].JoyState.RUp;
			case 44:return !mmJoystick.m_Joy[i].JoyState.RDown;
			case 45:return !mmJoystick.m_Joy[i].JoyState.UUp;
			case 46:return !mmJoystick.m_Joy[i].JoyState.UDown;
			case 47:return !mmJoystick.m_Joy[i].JoyState.VUp;
			case 48:return !mmJoystick.m_Joy[i].JoyState.VDown;

			default:
				if ((KeyIdent & 0xff) > 40)
					return false;
				return !mmJoystick.m_Joy[i].JoyState.Button [(KeyIdent & 0xff) - 8];
		}
	}
	return ((GetKeyState (KeyIdent) & 0x80) == 0);
}

static void S9xWinScanmmJoypads ()
{
	unsigned char PadState[2];
	int i;
	for (i = 0; i < NUM_JOYSTICKS; i++)
	{
		if (mmJoystick.m_Joy[i].m_bUseJoystick)
		{
			if (joyGetPosEx(mmJoystick.m_Joy[i].m_uJoyID, &mmJoystick.m_Joy[i].m_JoyInfo) != JOYERR_NOERROR)
			{
				mmJoystick.m_Joy[i].m_bUseJoystick = false;
				continue;
			}
			
			for (unsigned long B = 0; B < 32; B ++) {
				mmJoystick.m_Joy[i].JoyState.Button[B] = (mmJoystick.m_Joy[i].m_JoyInfo.dwButtons & (1 << B)) != 0;
			}
			if (mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_HASPOV
			&&  mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_POV4DIR)
			{
				switch (mmJoystick.m_Joy[i].m_JoyInfo.dwPOV)
				{
					case JOY_POVFORWARD:
						mmJoystick.m_Joy[i].JoyState.PovUp = true;
						mmJoystick.m_Joy[i].JoyState.PovRight = false;
						mmJoystick.m_Joy[i].JoyState.PovDown = false;
						mmJoystick.m_Joy[i].JoyState.PovLeft = false;
						break;

					case 4500:
						mmJoystick.m_Joy[i].JoyState.PovUp = true;
						mmJoystick.m_Joy[i].JoyState.PovRight = true;
						mmJoystick.m_Joy[i].JoyState.PovDown = false;
						mmJoystick.m_Joy[i].JoyState.PovLeft = false;
						break;

					case JOY_POVRIGHT:
						mmJoystick.m_Joy[i].JoyState.PovUp = false;
						mmJoystick.m_Joy[i].JoyState.PovRight = true;
						mmJoystick.m_Joy[i].JoyState.PovDown = false;
						mmJoystick.m_Joy[i].JoyState.PovLeft = false;
						break;

					case 13500:
						mmJoystick.m_Joy[i].JoyState.PovUp = false;
						mmJoystick.m_Joy[i].JoyState.PovRight = true;
						mmJoystick.m_Joy[i].JoyState.PovDown = true;
						mmJoystick.m_Joy[i].JoyState.PovLeft = false;
						break;

					case JOY_POVBACKWARD:
						mmJoystick.m_Joy[i].JoyState.PovUp = false;
						mmJoystick.m_Joy[i].JoyState.PovRight = false;
						mmJoystick.m_Joy[i].JoyState.PovDown = true;
						mmJoystick.m_Joy[i].JoyState.PovLeft = false;
						break;

					case 22500:
						mmJoystick.m_Joy[i].JoyState.PovUp = false;
						mmJoystick.m_Joy[i].JoyState.PovRight = false;
						mmJoystick.m_Joy[i].JoyState.PovDown = true;
						mmJoystick.m_Joy[i].JoyState.PovLeft = true;
						break;

					case JOY_POVLEFT:
						mmJoystick.m_Joy[i].JoyState.PovUp = false;
						mmJoystick.m_Joy[i].JoyState.PovRight = false;
						mmJoystick.m_Joy[i].JoyState.PovDown = false;
						mmJoystick.m_Joy[i].JoyState.PovLeft = true;
						break;

					case 31500:
						mmJoystick.m_Joy[i].JoyState.PovUp = true;
						mmJoystick.m_Joy[i].JoyState.PovRight = false;
						mmJoystick.m_Joy[i].JoyState.PovDown = false;
						mmJoystick.m_Joy[i].JoyState.PovLeft = true;
						break;

					default:
						mmJoystick.m_Joy[i].JoyState.PovUp = false;
						mmJoystick.m_Joy[i].JoyState.PovRight = false;
						mmJoystick.m_Joy[i].JoyState.PovDown = false;
						mmJoystick.m_Joy[i].JoyState.PovLeft = false;
						break;
				}
			}
			else
				mmJoystick.m_Joy[i].JoyState.PovUp = mmJoystick.m_Joy[i].JoyState.PovRight = mmJoystick.m_Joy[i].JoyState.PovDown = mmJoystick.m_Joy[i].JoyState.PovLeft = FALSE;
			mmJoystick.m_Joy[i].JoyState.Left = (mmJoystick.m_Joy[i].m_JoyInfo.dwXpos <= mmJoystick.m_Joy[i].m_dwXDZmin &&
												mmJoystick.m_Joy[i].m_dwXDZmin != mmJoystick.m_Joy[i].m_dwXDZmax);
			mmJoystick.m_Joy[i].JoyState.Right = (mmJoystick.m_Joy[i].m_dwXDZmax <= mmJoystick.m_Joy[i].m_JoyInfo.dwXpos &&
												mmJoystick.m_Joy[i].m_dwXDZmin != mmJoystick.m_Joy[i].m_dwXDZmax);
			mmJoystick.m_Joy[i].JoyState.Up = (mmJoystick.m_Joy[i].m_JoyInfo.dwYpos <= mmJoystick.m_Joy[i].m_dwYDZmin &&
												mmJoystick.m_Joy[i].m_dwYDZmin != mmJoystick.m_Joy[i].m_dwYDZmax);
			mmJoystick.m_Joy[i].JoyState.Down = (mmJoystick.m_Joy[i].m_dwYDZmax <= mmJoystick.m_Joy[i].m_JoyInfo.dwYpos &&
												mmJoystick.m_Joy[i].m_dwYDZmin != mmJoystick.m_Joy[i].m_dwYDZmax);
			if (mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_HASZ)
			{
				mmJoystick.m_Joy[i].JoyState.ZUp = (mmJoystick.m_Joy[i].m_JoyInfo.dwZpos <= mmJoystick.m_Joy[i].m_dwZDZmin &&
													mmJoystick.m_Joy[i].m_dwZDZmin != mmJoystick.m_Joy[i].m_dwZDZmax);
				mmJoystick.m_Joy[i].JoyState.ZDown = (mmJoystick.m_Joy[i].m_dwZDZmax <= mmJoystick.m_Joy[i].m_JoyInfo.dwZpos &&
													mmJoystick.m_Joy[i].m_dwZDZmin != mmJoystick.m_Joy[i].m_dwZDZmax);
			}
			else
				mmJoystick.m_Joy[i].JoyState.ZUp = mmJoystick.m_Joy[i].JoyState.ZDown = FALSE;
			if (mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_HASR)
			{
				mmJoystick.m_Joy[i].JoyState.RUp = (mmJoystick.m_Joy[i].m_JoyInfo.dwRpos <= mmJoystick.m_Joy[i].m_dwRDZmin &&
													mmJoystick.m_Joy[i].m_dwRDZmin != mmJoystick.m_Joy[i].m_dwRDZmax);
				mmJoystick.m_Joy[i].JoyState.RDown = (mmJoystick.m_Joy[i].m_dwRDZmax <= mmJoystick.m_Joy[i].m_JoyInfo.dwRpos &&
													mmJoystick.m_Joy[i].m_dwRDZmin != mmJoystick.m_Joy[i].m_dwRDZmax);
			}
			else
				mmJoystick.m_Joy[i].JoyState.RUp = mmJoystick.m_Joy[i].JoyState.RDown = FALSE;
			if (mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_HASU)
			{
				mmJoystick.m_Joy[i].JoyState.UUp = (mmJoystick.m_Joy[i].m_JoyInfo.dwUpos <= mmJoystick.m_Joy[i].m_dwUDZmin &&
													mmJoystick.m_Joy[i].m_dwUDZmin != mmJoystick.m_Joy[i].m_dwUDZmax);
				mmJoystick.m_Joy[i].JoyState.UDown = (mmJoystick.m_Joy[i].m_dwUDZmax <= mmJoystick.m_Joy[i].m_JoyInfo.dwUpos &&
													mmJoystick.m_Joy[i].m_dwUDZmin != mmJoystick.m_Joy[i].m_dwUDZmax);
			}
			else
				mmJoystick.m_Joy[i].JoyState.UUp = mmJoystick.m_Joy[i].JoyState.UDown = FALSE;
			if (mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_HASV)
			{
				mmJoystick.m_Joy[i].JoyState.VUp = (mmJoystick.m_Joy[i].m_JoyInfo.dwVpos <= mmJoystick.m_Joy[i].m_dwVDZmin &&
													mmJoystick.m_Joy[i].m_dwVDZmin != mmJoystick.m_Joy[i].m_dwVDZmax);
				mmJoystick.m_Joy[i].JoyState.VDown = (mmJoystick.m_Joy[i].m_dwVDZmax <= mmJoystick.m_Joy[i].m_JoyInfo.dwVpos &&
													mmJoystick.m_Joy[i].m_dwVDZmin != mmJoystick.m_Joy[i].m_dwVDZmax);
			}
			else
				mmJoystick.m_Joy[i].JoyState.VUp = mmJoystick.m_Joy[i].JoyState.VDown = FALSE;
		}
	}

	for (int J = 0; J < 5; J++)
	{
		if (Joypad [J].Enabled)
		{
			PadState[0]  = 0;
			PadState[0] |= !S9xGetStatemm (Joypad[J].R)      ?  16 : 0;
			PadState[0] |= !S9xGetStatemm (Joypad[J].L)      ?  32 : 0;
			PadState[0] |= !S9xGetStatemm (Joypad[J].X)      ?  64 : 0;
			PadState[0] |= !S9xGetStatemm (Joypad[J].A)      ? 128 : 0;

			PadState[1]  = 0;
			PadState[1] |= !S9xGetStatemm (Joypad[J].Right)  ?   1 : 0;
			PadState[1] |= !S9xGetStatemm (Joypad[J].Right_Up)  ? 1 + 8 : 0;
			PadState[1] |= !S9xGetStatemm (Joypad[J].Right_Down)? 1 + 4 : 0;
			PadState[1] |= !S9xGetStatemm (Joypad[J].Left)   ?   2 : 0;
			PadState[1] |= !S9xGetStatemm (Joypad[J].Left_Up)?   2 + 8 : 0;
			PadState[1] |= !S9xGetStatemm (Joypad[J].Left_Down)?  2 + 4 : 0;
			PadState[1] |= !S9xGetStatemm (Joypad[J].Down)   ?   4 : 0;
			PadState[1] |= !S9xGetStatemm (Joypad[J].Up)     ?   8 : 0;
			PadState[1] |= !S9xGetStatemm (Joypad[J].Start)  ?  16 : 0;
			PadState[1] |= !S9xGetStatemm (Joypad[J].Select) ?  32 : 0;
			PadState[1] |= !S9xGetStatemm (Joypad[J].Y)      ?  64 : 0;
			PadState[1] |= !S9xGetStatemm (Joypad[J].B)      ? 128 : 0;
			joypads [J] = PadState [0] | (PadState [1] << 8) | 0x80000000;
		}
		else
			joypads [J] = 0;
	}

	MapFuncPress = FALSE;
	for(i = 0; i < 8; i++) {
		if(!S9xGetStatemm(LOWORD(MapFunc[i]))) {
			if(!MapFuncPress) {
				MapFuncPress = TRUE;
				if(!MapFuncPressed) {
					MapFuncPressed = MapFuncPress;
					TurbofwdPressed[i] = (HIWORD(MapFunc[i]) == MF_TURBOPAUSEFWD);
					PostMessage (GUI.hWnd, WM_COMMAND, ID_MAPFUNC, HIWORD(MapFunc[i]));
					return;
				}
			}
		}
		else if(TurbofwdPressed[i] && HIWORD(MapFunc[i]) == MF_TURBOPAUSEFWD) {
			TurbofwdPressed[i] = FALSE;
			MapFuncPressed = MapFuncPress;
			PostMessage (GUI.hWnd, WM_COMMAND, ID_MAPFUNC, HIWORD(MapFunc[i]));
			return;
		}
	}
	MapFuncPressed = MapFuncPress;
}

static unsigned long TurboStatus [5] = {~0, ~0, ~0, ~0, ~0};
static unsigned long JoyReadCount;

extern "C" unsigned int S9xReadJoypad (int which1)
{
       if(JoyReadCount>=9) JoyReadCount=0; // Original
	++JoyReadCount;
	if(GetKeyState(VK_CONTROL)&0x8000 || GetKeyState(VK_MENU)&0x8000 || GetKeyState(VK_SHIFT)&0x8000)
		return 0;

	if (which1 > 4)
		return 0;

	if(JoyReadCount & Options.JoyReadRateMask) {
		if(JoyReadCount & 1)
			return (joypads [which1]);
		else
			return (joypads [which1] & TurboStatus[which1]);
	}

	if (which1 == 0/* && !Settings.NetPlay*/)
		S9xWinScanJoypads ();

#ifdef NETPLAY_SUPPORT
	if (Settings.NetPlay) {
		if(JoyReadCount & 1)
			return (S9xNPGetJoypad (which1));
		else
			return (S9xNPGetJoypad (which1) & TurboStatus[which1]);
	}
#endif

	if(JoyReadCount & 1)
		return (joypads [which1]);
	else
		return (joypads [which1] & TurboStatus[which1]);
}

void MakeTurboStatus()
{
	unsigned char PadState[2];
	for (int J = 0; J < 5; J++)
	{
		if (Joypad [J].Enabled)
		{
			TurboStatus[J] = ~0;
			PadState[0]  = 0;
			PadState[0] |= (Joypad[J].Turbo_R ? 16 : 0);
			PadState[0] |= (Joypad[J].Turbo_L ? 32 : 0);
			PadState[0] |= (Joypad[J].Turbo_X ? 64 : 0);
			PadState[0] |= (Joypad[J].Turbo_A ? 128 : 0);

			PadState[1]  = 0;
			PadState[1] |= (Joypad[J].Turbo_Right ? 1 : 0);
			PadState[1] |= (Joypad[J].Turbo_Left ? 2 : 0);
			PadState[1] |= (Joypad[J].Turbo_Down ? 4 : 0);
			PadState[1] |= (Joypad[J].Turbo_Up ? 8 : 0);
			PadState[1] |= (Joypad[J].Turbo_Start ? 16 : 0);
			PadState[1] |= (Joypad[J].Turbo_Select ? 32 : 0);
			PadState[1] |= (Joypad[J].Turbo_Y ? 64 : 0);
			PadState[1] |= (Joypad[J].Turbo_B ? 128 : 0);
			TurboStatus[J] ^= (PadState [0] | (PadState [1] << 8));
		}
		else
			TurboStatus[J] = ~0;
	}
}

static WORD GetmmJoyButtonCode()
{
	static WORD mjb;
	static WORD mB;
	static bool mJoyButtonDownFlag;
	static bool mJoyPovDownFlag;
	static bool mJoyAxisxDownFlag;
	static bool mJoyAxisyDownFlag;
	static bool mJoyAxiszDownFlag;
	static bool mJoyAxisrDownFlag;
	static bool mJoyAxisuDownFlag;
	static bool mJoyAxisvDownFlag;
	static bool mJoyButtonDownedFlag;
	static bool mJoyPovDownedFlag;
	static bool mJoyAxisxDownedFlag;
	static bool mJoyAxisyDownedFlag;
	static bool mJoyAxiszDownedFlag;
	static bool mJoyAxisrDownedFlag;
	static bool mJoyAxisuDownedFlag;
	static bool mJoyAxisvDownedFlag;
	static unsigned long mJoyButtonCheck;

	mjb = 0;
	mJoyButtonDownFlag = mJoyPovDownFlag = mJoyAxisxDownFlag = mJoyAxisyDownFlag = mJoyAxiszDownFlag = mJoyAxisrDownFlag = mJoyAxisuDownFlag = mJoyAxisvDownFlag = 0;
	++mJoyButtonCheck;
	for(WORD i = 0; i < NUM_JOYSTICKS; i++) {
		if(mmJoystick.m_Joy[i].m_bUseJoystick && (mJoyButtonCheck % 8) == 0)
		{
			if(joyGetPosEx(mmJoystick.m_Joy[i].m_uJoyID, &mmJoystick.m_Joy[i].m_JoyInfo) != JOYERR_NOERROR)
			{
				mmJoystick.m_Joy[i].m_bUseJoystick = FALSE;
				continue;
			}
			for(mB = 0; mB < 32; mB ++) {
				if(mmJoystick.m_Joy[i].m_JoyInfo.dwButtons & (1 << mB))
				{
					mJoyButtonDownFlag = TRUE;
					if(mJoyButtonDownedFlag)
						break;
					mjb = ((i << 8) | 0x8000 | (mB + 8));
					break;
				}
			}
			if(mJoyButtonDownFlag)
				break;
		}
		else if(mmJoystick.m_Joy[i].m_bUseJoystick && (mJoyButtonCheck % 8) == 1 &&
				mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_HASPOV &&
				mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_POV4DIR)
		{
			if(joyGetPosEx(mmJoystick.m_Joy[i].m_uJoyID, &mmJoystick.m_Joy[i].m_JoyInfo) != JOYERR_NOERROR)
			{
				mmJoystick.m_Joy[i].m_bUseJoystick = FALSE;
				continue;
			}

			if(mmJoystick.m_Joy[i].m_JoyInfo.dwPOV == JOY_POVLEFT)
			{
				mJoyPovDownFlag = TRUE;
				if(mJoyPovDownedFlag)
					break;
				mjb = ((i << 8) | 0x8000 | 4);
				break;
			}

			else if(mmJoystick.m_Joy[i].m_JoyInfo.dwPOV == JOY_POVRIGHT)
			{
				mJoyPovDownFlag = TRUE;
				if(mJoyPovDownedFlag)
					break;
				mjb = ((i << 8) | 0x8000 | 5);
				break;
			}

			else if(mmJoystick.m_Joy[i].m_JoyInfo.dwPOV == JOY_POVFORWARD)
			{
				mJoyPovDownFlag = TRUE;
				if(mJoyPovDownedFlag)
					break;
				mjb = ((i << 8) | 0x8000 | 6);
				break;
			}

			else if(mmJoystick.m_Joy[i].m_JoyInfo.dwPOV == JOY_POVBACKWARD)
			{
				mJoyPovDownFlag = TRUE;
				if(mJoyPovDownedFlag)
					break;
				mjb = ((i << 8) | 0x8000 | 7);
				break;
			}
		}
		else if(mmJoystick.m_Joy[i].m_bUseJoystick && (mJoyButtonCheck % 8) == 2)
		{
			if(joyGetPosEx(mmJoystick.m_Joy[i].m_uJoyID, &mmJoystick.m_Joy[i].m_JoyInfo) != JOYERR_NOERROR)
			{
				mmJoystick.m_Joy[i].m_bUseJoystick = FALSE;
				continue;
			}
			//Left
			if(mmJoystick.m_Joy[i].m_JoyInfo.dwXpos <= mmJoystick.m_Joy[i].m_dwXDZmin && mmJoystick.m_Joy[i].m_dwXDZmin != mmJoystick.m_Joy[i].m_dwXDZmax)
			{
				mJoyAxisxDownFlag = TRUE;
				if(mJoyAxisxDownedFlag)
					break;
				mjb = ((i << 8) | 0x8000 | 0);
				break;
			}
			//Right
			else if(mmJoystick.m_Joy[i].m_dwXDZmax <= mmJoystick.m_Joy[i].m_JoyInfo.dwXpos && mmJoystick.m_Joy[i].m_dwXDZmin != mmJoystick.m_Joy[i].m_dwXDZmax)
			{
				mJoyAxisxDownFlag = TRUE;
				if(mJoyAxisxDownedFlag)
					break;
				mjb = ((i << 8) | 0x8000 | 1);
				break;
			}
		}
		else if(mmJoystick.m_Joy[i].m_bUseJoystick && (mJoyButtonCheck % 8) == 3)
		{
			if(joyGetPosEx(mmJoystick.m_Joy[i].m_uJoyID, &mmJoystick.m_Joy[i].m_JoyInfo) != JOYERR_NOERROR)
			{
				mmJoystick.m_Joy[i].m_bUseJoystick = FALSE;
				continue;
			}
			//Up
			if(mmJoystick.m_Joy[i].m_JoyInfo.dwYpos <= mmJoystick.m_Joy[i].m_dwYDZmin && mmJoystick.m_Joy[i].m_dwYDZmin != mmJoystick.m_Joy[i].m_dwYDZmax)
			{
				mJoyAxisyDownFlag = TRUE;
				if(mJoyAxisyDownedFlag)
					break;
				mjb = ((i << 8) | 0x8000 | 2);
				break;
			}
			//Down
			else if(mmJoystick.m_Joy[i].m_dwYDZmax <= mmJoystick.m_Joy[i].m_JoyInfo.dwYpos && mmJoystick.m_Joy[i].m_dwYDZmin != mmJoystick.m_Joy[i].m_dwYDZmax)
			{
				mJoyAxisyDownFlag = TRUE;
				if(mJoyAxisyDownedFlag)
					break;
				mjb = ((i << 8) | 0x8000 | 3);
				break;
			}
		}
		else if(mmJoystick.m_Joy[i].m_bUseJoystick && (mJoyButtonCheck % 8) == 4)
		{
			if(joyGetPosEx(mmJoystick.m_Joy[i].m_uJoyID, &mmJoystick.m_Joy[i].m_JoyInfo) != JOYERR_NOERROR)
			{
				mmJoystick.m_Joy[i].m_bUseJoystick = FALSE;
				continue;
			}
			if (mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_HASZ)
			{
				//ZUp
				if(mmJoystick.m_Joy[i].m_JoyInfo.dwZpos <= mmJoystick.m_Joy[i].m_dwZDZmin && mmJoystick.m_Joy[i].m_dwZDZmin != mmJoystick.m_Joy[i].m_dwZDZmax)
				{
					mJoyAxiszDownFlag = TRUE;
					if(mJoyAxiszDownedFlag)
						break;
					mjb = ((i << 8) | 0x8000 | 41);
					break;
				}
				//ZDown
				else if(mmJoystick.m_Joy[i].m_dwZDZmax <= mmJoystick.m_Joy[i].m_JoyInfo.dwZpos && mmJoystick.m_Joy[i].m_dwZDZmin != mmJoystick.m_Joy[i].m_dwZDZmax)
				{
					mJoyAxiszDownFlag = TRUE;
					if(mJoyAxiszDownedFlag)
						break;
					mjb = ((i << 8) | 0x8000 | 42);
					break;
				}
			}
		}
		else if(mmJoystick.m_Joy[i].m_bUseJoystick && (mJoyButtonCheck % 8) == 5)
		{
			if(joyGetPosEx(mmJoystick.m_Joy[i].m_uJoyID, &mmJoystick.m_Joy[i].m_JoyInfo) != JOYERR_NOERROR)
			{
				mmJoystick.m_Joy[i].m_bUseJoystick = FALSE;
				continue;
			}
			if (mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_HASR)
			{
				//RUp
				if(mmJoystick.m_Joy[i].m_JoyInfo.dwRpos <= mmJoystick.m_Joy[i].m_dwRDZmin && mmJoystick.m_Joy[i].m_dwRDZmin != mmJoystick.m_Joy[i].m_dwRDZmax)
				{
					mJoyAxisrDownFlag = TRUE;
					if(mJoyAxisrDownedFlag)
						break;
					mjb = ((i << 8) | 0x8000 | 43);
					break;
				}
				//RDown
				else if(mmJoystick.m_Joy[i].m_dwRDZmax <= mmJoystick.m_Joy[i].m_JoyInfo.dwRpos && mmJoystick.m_Joy[i].m_dwRDZmin != mmJoystick.m_Joy[i].m_dwRDZmax)
				{
					mJoyAxisrDownFlag = TRUE;
					if(mJoyAxisrDownedFlag)
						break;
					mjb = ((i << 8) | 0x8000 | 44);
					break;
				}
			}
		}
		else if(mmJoystick.m_Joy[i].m_bUseJoystick && (mJoyButtonCheck % 8) == 6)
		{
			if(joyGetPosEx(mmJoystick.m_Joy[i].m_uJoyID, &mmJoystick.m_Joy[i].m_JoyInfo) != JOYERR_NOERROR)
			{
				mmJoystick.m_Joy[i].m_bUseJoystick = FALSE;
				continue;
			}
			if (mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_HASU)
			{
				//UUp
				if(mmJoystick.m_Joy[i].m_JoyInfo.dwUpos <= mmJoystick.m_Joy[i].m_dwUDZmin && mmJoystick.m_Joy[i].m_dwUDZmin != mmJoystick.m_Joy[i].m_dwUDZmax)
				{
					mJoyAxisuDownFlag = TRUE;
					if(mJoyAxisuDownedFlag)
						break;
					mjb = ((i << 8) | 0x8000 | 45);
					break;
				}
				//UDown
				else if(mmJoystick.m_Joy[i].m_dwUDZmax <= mmJoystick.m_Joy[i].m_JoyInfo.dwUpos && mmJoystick.m_Joy[i].m_dwUDZmin != mmJoystick.m_Joy[i].m_dwUDZmax)
				{
					mJoyAxisuDownFlag = TRUE;
					if(mJoyAxisuDownedFlag)
						break;
					mjb = ((i << 8) | 0x8000 | 46);
					break;
				}
			}
		}
		else if(mmJoystick.m_Joy[i].m_bUseJoystick && (mJoyButtonCheck % 8) == 7)
		{
			if(joyGetPosEx(mmJoystick.m_Joy[i].m_uJoyID, &mmJoystick.m_Joy[i].m_JoyInfo) != JOYERR_NOERROR)
			{
				mmJoystick.m_Joy[i].m_bUseJoystick = FALSE;
				continue;
			}
			if (mmJoystick.m_Joy[i].m_JoyCaps.wCaps & JOYCAPS_HASV)
			{
				//VUp
				if(mmJoystick.m_Joy[i].m_JoyInfo.dwVpos <= mmJoystick.m_Joy[i].m_dwVDZmin && mmJoystick.m_Joy[i].m_dwVDZmin != mmJoystick.m_Joy[i].m_dwVDZmax)
				{
					mJoyAxisvDownFlag = TRUE;
					if(mJoyAxisvDownedFlag)
						break;
					mjb = ((i << 8) | 0x8000 | 47);
					break;
				}
				//VDown
				else if(mmJoystick.m_Joy[i].m_dwVDZmax <= mmJoystick.m_Joy[i].m_JoyInfo.dwVpos && mmJoystick.m_Joy[i].m_dwVDZmin != mmJoystick.m_Joy[i].m_dwVDZmax)
				{
					mJoyAxisvDownFlag = TRUE;
					if(mJoyAxisvDownedFlag)
						break;
					mjb = ((i << 8) | 0x8000 | 48);
					break;
				}
			}
		}
	}
	if(mJoyButtonDownFlag)
		mJoyButtonDownedFlag = mJoyButtonDownFlag;
	else if(mJoyPovDownFlag)
		mJoyPovDownedFlag = mJoyPovDownFlag;
	else if(mJoyAxisxDownFlag)
		mJoyAxisxDownedFlag = mJoyAxisxDownFlag;
	else if(mJoyAxisyDownFlag)
		mJoyAxisyDownedFlag = mJoyAxisyDownFlag;
	else if(mJoyAxiszDownFlag)
		mJoyAxiszDownedFlag = mJoyAxiszDownFlag;
	else if(mJoyAxisrDownFlag)
		mJoyAxisrDownedFlag = mJoyAxisrDownFlag;
	else if(mJoyAxisuDownFlag)
		mJoyAxisuDownedFlag = mJoyAxisuDownFlag;
	else if(mJoyAxisvDownFlag)
		mJoyAxisvDownedFlag = mJoyAxisvDownFlag;
	if(!mJoyButtonDownFlag && (mJoyButtonCheck % 8) == 0)
		mJoyButtonDownedFlag = 0;
	else if(!mJoyPovDownFlag && (mJoyButtonCheck % 8) == 1)
		mJoyPovDownedFlag = 0;
	else if(!mJoyAxisxDownFlag && (mJoyButtonCheck % 8) == 2)
		mJoyAxisxDownedFlag = 0;
	else if(!mJoyAxisyDownFlag && (mJoyButtonCheck % 8) == 3)
		mJoyAxisyDownedFlag = 0;
	else if(!mJoyAxiszDownFlag && (mJoyButtonCheck % 8) == 4)
		mJoyAxiszDownedFlag = 0;
	else if(!mJoyAxisrDownFlag && (mJoyButtonCheck % 8) == 5)
		mJoyAxisrDownedFlag = 0;
	else if(!mJoyAxisuDownFlag && (mJoyButtonCheck % 8) == 6)
		mJoyAxisuDownedFlag = 0;
	else if(!mJoyAxisvDownFlag && (mJoyButtonCheck % 8) == 7)
		mJoyAxisvDownedFlag = 0;
	return mjb;
}

static BOOL CALLBACK DIJoystick_EnumDeviceProc(LPDIDEVICEINSTANCE pdidi, LPVOID pv)
{
	if(!lstrcmp(pdidi->tszProductName, "ACT LABS GS") && !lstrcmp(pdidi->tszInstanceName, "ACT LABS GS"))
		return DIENUM_CONTINUE;
	diJoystick.joysticks[diJoystick.num_joysticks].guidDevice = pdidi->guidInstance;
	diJoystick.num_joysticks++;
	return DIENUM_CONTINUE;
}

static HRESULT SetDIDwordProperty(LPDIRECTINPUTDEVICE2 pdev, REFGUID guidProperty,
               DWORD dwObject, DWORD dwHow, DWORD dwValue)
{
    DIPROPDWORD dipdw;

    dipdw.diph.dwSize       = sizeof(dipdw);
    dipdw.diph.dwHeaderSize = sizeof(dipdw.diph);
    dipdw.diph.dwObj        = dwObject;
    dipdw.diph.dwHow        = dwHow;
    dipdw.dwData            = dwValue;

    return pdev->SetProperty(guidProperty, &dipdw.diph);

}

static BOOL CALLBACK DIJoystick_EnumAxisObjectsProc(LPCDIDEVICEOBJECTINSTANCE lpddoi,
                                             dijoystick_type *joystick)
{
	DIPROPRANGE diprg;
	HRESULT hr;

	joystick->axes[joystick->num_axes].guid = lpddoi->guidType;

	joystick->axes[joystick->num_axes].offset = lpddoi->dwOfs;

	//ErrorMsg("got axis %s, offset %i",lpddoi->tszName, lpddoi->dwOfs);
	diprg.diph.dwSize       = sizeof(diprg);
	diprg.diph.dwHeaderSize = sizeof(diprg.diph);
	diprg.diph.dwObj        = lpddoi->dwOfs;
	diprg.diph.dwHow        = DIPH_BYOFFSET;
	diprg.lMin              = 0;
	diprg.lMax              = 255;

	hr = joystick->did->SetProperty(DIPROP_RANGE, &diprg.diph);
	if (FAILED(hr)) /* if this fails, don't use this axis */
		return DIENUM_CONTINUE;

	/* Set axis dead zone to 0; we need accurate #'s for analog joystick reading. */
	hr = SetDIDwordProperty(joystick->did, DIPROP_DEADZONE, lpddoi->dwOfs, DIPH_BYOFFSET, 0);

	joystick->num_axes++;

	return DIENUM_CONTINUE;
}

static BOOL CALLBACK DIJoystick_EnumPOVObjectsProc(LPCDIDEVICEOBJECTINSTANCE lpddoi,
                                            dijoystick_type *joystick)
{
   joystick->num_pov++;
   
   return DIENUM_CONTINUE;
}

static BOOL CALLBACK DIJoystick_EnumButtonObjectsProc(LPCDIDEVICEOBJECTINSTANCE lpddoi,
                                               dijoystick_type *joystick)
{
   joystick->num_buttons++;
   
   return DIENUM_CONTINUE;
}

static void ClearJoyState(DIJOYSTATE *pdijs)
{
    memset(pdijs,0,sizeof(DIJOYSTATE));
    pdijs->lX = 128;
    pdijs->lY = 128;
    pdijs->lZ = 128;
    pdijs->lRx = 128;
    pdijs->lRy = 128;
    pdijs->lRz = 128;
    pdijs->rglSlider[0] = 128;
    pdijs->rglSlider[1] = 128;
    pdijs->rgdwPOV[0] = (DWORD)-1;
    pdijs->rgdwPOV[1] = (DWORD)-1;
    pdijs->rgdwPOV[2] = (DWORD)-1;
    pdijs->rgdwPOV[3] = (DWORD)-1;
}

static void dInitJoystick(dijoystick_type *joystick)
{
	LPDIRECTINPUTDEVICE didTemp;
	HRESULT hr;

	joystick->use_joystick = FALSE;

	joystick->did = NULL;
	joystick->num_axes = 0;

	/* get a did1 interface first... */
	hr = DirectX.lpDI->CreateDevice(joystick->guidDevice, &didTemp, NULL);
	if (FAILED(hr))
		return;

	/* get a did2 interface to work with polling (most) joysticks */
	hr = didTemp->QueryInterface(IID_IDirectInputDevice2, (void **)&joystick->did);

	/* dispose of the temp interface */
	didTemp->Release();

	/* check result of getting the did2 */
	if (FAILED(hr))
	{
		/* no error message because this happens in dx3 */
		/* ErrorMsg("DirectInput QueryInterface joystick failed\n"); */
		joystick->did = NULL;
		return;
	}

	hr = joystick->did->SetCooperativeLevel(DirectX.hWnd, DISCL_NONEXCLUSIVE | DISCL_BACKGROUND);
	if (FAILED(hr))
		return;

	hr = joystick->did->SetDataFormat(&c_dfDIJoystick);
	if (FAILED(hr))
		return;

	/* enumerate our axes */
	hr = joystick->did->EnumObjects((LPDIENUMDEVICEOBJECTSCALLBACK)DIJoystick_EnumAxisObjectsProc, joystick, DIDFT_AXIS);

	if (FAILED(hr))
		return;


	 /* enumerate our POV hats */
	joystick->num_pov = 0;
	hr = joystick->did->EnumObjects((LPDIENUMDEVICEOBJECTSCALLBACK)DIJoystick_EnumPOVObjectsProc, joystick, DIDFT_POV);
	if (FAILED(hr))
		return;

	/* enumerate our buttons */

	joystick->num_buttons = 0;
	hr = joystick->did->EnumObjects((LPDIENUMDEVICEOBJECTSCALLBACK)DIJoystick_EnumButtonObjectsProc, joystick, DIDFT_BUTTON);
	if (FAILED(hr))
		return;

	hr = joystick->did->Acquire();
	if (FAILED(hr)) 
		return;

	/* start by clearing the structures */

	ClearJoyState(&joystick->dijs);

	joystick->use_joystick = TRUE;
}

static unsigned char DIJoystick_init()
{
	DWORD i;
	HRESULT hr;

	diJoystick.num_joysticks = 0;

	/* enumerate for joystick devices */
	hr = DirectX.lpDI->EnumDevices(DIDEVTYPE_JOYSTICK,
								(LPDIENUMDEVICESCALLBACK)DIJoystick_EnumDeviceProc,
								NULL,
								DIEDFL_ALLDEVICES);
	if (FAILED(hr))
		return FALSE;

	/* create each joystick device, enumerate each joystick for axes, etc */
	for (i = 0; i < diJoystick.num_joysticks; i++)
	{
		dInitJoystick(&diJoystick.joysticks[i]);
	}
	return TRUE;
}

#if 0
static unsigned char DIKeyboard_init()
{
	HRESULT hr;
	int i;

	diKeyboard.m_didKeyboard  = NULL;

	for (i=0;i<256;i++)
		diKeyboard.m_key[i] = 0;


	if (DirectX.lpDI == NULL)
		return FALSE;

	/* setup the keyboard */
	hr = DirectX.lpDI->CreateDevice(GUID_SysKeyboard, &diKeyboard.m_didKeyboard, NULL);

	if (FAILED(hr)) 
		return FALSE;

	hr = diKeyboard.m_didKeyboard->SetDataFormat(&c_dfDIKeyboard);

	if (FAILED(hr)) 
		return FALSE;

	hr = diKeyboard.m_didKeyboard->SetCooperativeLevel(DirectX.hWnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);

	if (FAILED(hr)) 
		return FALSE;

	hr = diKeyboard.m_didKeyboard->Acquire();

	Mkvk2diktable();
	return TRUE;
}
#endif

#if 0
static unsigned char DIMouse_init()
{
	HRESULT hr;

	diMouse.m_didMouse  = NULL;

	if (DirectX.lpDI == NULL)
		return FALSE;

	hr = DirectX.lpDI->CreateDevice(GUID_SysMouse, &diMouse.m_didMouse, NULL);

	if (FAILED(hr)) 
		return FALSE;

	hr = diMouse.m_didMouse->SetDataFormat(&c_dfDIMouse);

	if (FAILED(hr)) 
		return FALSE;

	hr = diMouse.m_didMouse->SetCooperativeLevel(DirectX.hWnd, DISCL_EXCLUSIVE | DISCL_FOREGROUND);

	if (FAILED(hr)) 
		return FALSE;

	hr = diMouse.m_didMouse->Acquire();

	return TRUE;
}
#endif

static unsigned char DirectInput_init()
{
	if(!DirectX_DirectInputInitialize() ||
		!DIJoystick_init()/* ||
		!DIKeyboard_init() ||
		!DIMouse_init()*/)
		return FALSE;
	return TRUE;
}

static void ExitdiJoystick(dijoystick_type *joystick)
{
	if (joystick->did != NULL)
	{
		joystick->did->Unacquire();
		joystick->did->Release();
		joystick->did = NULL;
	}
}

static void DIJoystick_exit()
{
	DWORD i;
	for (i = 0; i < diJoystick.num_joysticks; i++)
		ExitdiJoystick(&diJoystick.joysticks[i]);
	diJoystick.num_joysticks = 0;
}

#if 0
static void DIKeyboard_exit()
{
	if (!diKeyboard.m_didKeyboard)
		return;

	/*
	 Cleanliness is next to godliness.  Unacquire the device
	 one last time just in case we got really confused and tried
	 to exit while the device is still acquired.
	 */
	diKeyboard.m_didKeyboard->Unacquire();
	diKeyboard.m_didKeyboard->Release();
	diKeyboard.m_didKeyboard = NULL;
}
#endif

#if 0
static void DIMouse_exit()
{
	if (!diMouse.m_didMouse)
		return;

	diMouse.m_didMouse->Unacquire();
	diMouse.m_didMouse->Release();
	diMouse.m_didMouse = NULL;
}
#endif

void DirectInput_exit()
{
	DIJoystick_exit();
//	DIKeyboard_exit();
//	DIMouse_exit();
	DirectX_DirectInputDeInitialize();
}

static bool S9xGetStatedi (WORD KeyIdent)
{
	if (KeyIdent & 0x8000)
	{
		int i = (KeyIdent >> 8) & 15;

		switch (KeyIdent & 0xff)
		{
			case 0: return !diJoystick.joysticks[i].JoyState.Left;
			case 1: return !diJoystick.joysticks[i].JoyState.Right;
			case 2: return !diJoystick.joysticks[i].JoyState.Up;
			case 3: return !diJoystick.joysticks[i].JoyState.Down;
			case 4: return !diJoystick.joysticks[i].JoyState.PovLeft;
			case 5: return !diJoystick.joysticks[i].JoyState.PovRight;
			case 6: return !diJoystick.joysticks[i].JoyState.PovUp;
			case 7: return !diJoystick.joysticks[i].JoyState.PovDown;
			case 41:return !diJoystick.joysticks[i].JoyState.ZUp;
			case 42:return !diJoystick.joysticks[i].JoyState.ZDown;
			case 43:return !diJoystick.joysticks[i].JoyState.RUp;
			case 44:return !diJoystick.joysticks[i].JoyState.RDown;
			case 45:return !diJoystick.joysticks[i].JoyState.UUp;
			case 46:return !diJoystick.joysticks[i].JoyState.UDown;
			case 47:return !diJoystick.joysticks[i].JoyState.VUp;
			case 48:return !diJoystick.joysticks[i].JoyState.VDown;

			default:
				if ((KeyIdent & 0xff) > 40)
					return false;
				return !diJoystick.joysticks[i].JoyState.Button [(KeyIdent & 0xff) - 8];
		}
	}
	return ((GetKeyState (KeyIdent) & 0x80) == 0);
}

#define DIDZ 60

static void S9xWinScandiJoypads ()
{
	HRESULT hr;
	DWORD i, i2;
	unsigned char PadState[2];

	for (i = 0; i < diJoystick.num_joysticks; i++)
	{
		ClearJoyState(&diJoystick.joysticks[i].dijs);
		if (diJoystick.joysticks[i].did && diJoystick.joysticks[i].use_joystick)
		{
			hr = diJoystick.joysticks[i].did->Poll();
			hr = diJoystick.joysticks[i].did->GetDeviceState(sizeof(DIJOYSTATE), &diJoystick.joysticks[i].dijs);

			if (FAILED(hr))
			{
				if (hr == DIERR_INPUTLOST || hr == DIERR_NOTACQUIRED)
				{
					hr = diJoystick.joysticks[i].did->Acquire();
				}
				continue;
			}
			
			//button
			for (unsigned long B = 0; B < 32; B ++) {
				diJoystick.joysticks[i].JoyState.Button[B] = (diJoystick.joysticks[i].dijs.rgbButtons[B]  & 0x80) != 0;
			}

			//pov
			diJoystick.joysticks[i].JoyState.PovLeft = diJoystick.joysticks[i].JoyState.PovRight = diJoystick.joysticks[i].JoyState.PovUp = diJoystick.joysticks[i].JoyState.PovDown = FALSE;
			if (diJoystick.joysticks[i].num_pov)
			{
				int pov_value;
				int angle;
				int axis_value_x, axis_value_y;
				for(i2 = 0; i2 < diJoystick.joysticks[i].num_pov; i2++) {
					pov_value = (int)diJoystick.joysticks[i].dijs.rgdwPOV[i2];
					if(LOWORD(pov_value) != 0xffff) {
						angle = (pov_value + 27000) % 36000;
						angle = (36000 - angle) % 36000;
						angle /= 100;
						axis_value_x = 128 + (int)(127*cos(2*PI*angle/360.0)); // x
						axis_value_y = 128 + (int)(127*sin(2*PI*angle/360.0)); // y
						if(axis_value_x <= (128 - 128*DIDZ/100))
							diJoystick.joysticks[i].JoyState.PovLeft = TRUE;
						if(axis_value_x >= (128 + 128*DIDZ/100))
							diJoystick.joysticks[i].JoyState.PovRight = TRUE;
						if(axis_value_y <= (128 - 128*DIDZ/100))
							diJoystick.joysticks[i].JoyState.PovUp = TRUE;
						if(axis_value_y >= (128 + 128*DIDZ/100))
							diJoystick.joysticks[i].JoyState.PovDown = TRUE;
					}
				}
			}

			//axis
			//
			//X
			diJoystick.joysticks[i].JoyState.Left = (diJoystick.joysticks[i].num_axes >= 1 &&
				*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[0].offset) <= (128 - 128*DIDZ/100)) ? TRUE : FALSE;
			diJoystick.joysticks[i].JoyState.Right = (diJoystick.joysticks[i].num_axes >= 1 &&
				*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[0].offset) >= (128 + 128*DIDZ/100)) ? TRUE : FALSE;
			//Y
			diJoystick.joysticks[i].JoyState.Up = (diJoystick.joysticks[i].num_axes >= 2 &&
				*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[1].offset) <= (128 - 128*DIDZ/100)) ? TRUE : FALSE;
			diJoystick.joysticks[i].JoyState.Down = (diJoystick.joysticks[i].num_axes >= 2 &&
				*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[1].offset) >= (128 + 128*DIDZ/100)) ? TRUE : FALSE;
			//Z
			diJoystick.joysticks[i].JoyState.ZUp = (diJoystick.joysticks[i].num_axes >= 3 &&
				*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[2].offset) <= (128 - 128*DIDZ/100)) ? TRUE : FALSE;
			diJoystick.joysticks[i].JoyState.ZDown = (diJoystick.joysticks[i].num_axes >= 3 &&
				*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[2].offset) >= (128 + 128*DIDZ/100)) ? TRUE : FALSE;
			//R
			diJoystick.joysticks[i].JoyState.RUp = (diJoystick.joysticks[i].num_axes >= 4 &&
				*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[3].offset) <= (128 - 128*DIDZ/100)) ? TRUE : FALSE;
			diJoystick.joysticks[i].JoyState.RDown = (diJoystick.joysticks[i].num_axes >= 4 &&
				*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[3].offset) >= (128 + 128*DIDZ/100)) ? TRUE : FALSE;
			//U
			diJoystick.joysticks[i].JoyState.UUp = (diJoystick.joysticks[i].num_axes >= 5 &&
				*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[4].offset) <= (128 - 128*DIDZ/100)) ? TRUE : FALSE;
			diJoystick.joysticks[i].JoyState.UDown = (diJoystick.joysticks[i].num_axes >= 5 &&
				*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[4].offset) >= (128 + 128*DIDZ/100)) ? TRUE : FALSE;
			//V
			diJoystick.joysticks[i].JoyState.VUp = (diJoystick.joysticks[i].num_axes >= 6 &&
				*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[5].offset) <= (128 - 128*DIDZ/100)) ? TRUE : FALSE;
			diJoystick.joysticks[i].JoyState.VDown = (diJoystick.joysticks[i].num_axes >= 6 &&
				*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[5].offset) >= (128 + 128*DIDZ/100)) ? TRUE : FALSE;
		}
	}
	for (int J = 0; J < 5; J++)
	{
		if (Joypad [J].Enabled)
		{
			PadState[0]  = 0;
			PadState[0] |= !S9xGetStatedi (Joypad[J].R)      ?  16 : 0;
			PadState[0] |= !S9xGetStatedi (Joypad[J].L)      ?  32 : 0;
			PadState[0] |= !S9xGetStatedi (Joypad[J].X)      ?  64 : 0;
			PadState[0] |= !S9xGetStatedi (Joypad[J].A)      ? 128 : 0;

			PadState[1]  = 0;
			PadState[1] |= !S9xGetStatedi (Joypad[J].Right)  ?   1 : 0;
			PadState[1] |= !S9xGetStatedi (Joypad[J].Right_Up)  ? 1 + 8 : 0;
			PadState[1] |= !S9xGetStatedi (Joypad[J].Right_Down)? 1 + 4 : 0;
			PadState[1] |= !S9xGetStatedi (Joypad[J].Left)   ?   2 : 0;
			PadState[1] |= !S9xGetStatedi (Joypad[J].Left_Up)?   2 + 8 : 0;
			PadState[1] |= !S9xGetStatedi (Joypad[J].Left_Down)?  2 + 4 : 0;
			PadState[1] |= !S9xGetStatedi (Joypad[J].Down)   ?   4 : 0;
			PadState[1] |= !S9xGetStatedi (Joypad[J].Up)     ?   8 : 0;
			PadState[1] |= !S9xGetStatedi (Joypad[J].Start)  ?  16 : 0;
			PadState[1] |= !S9xGetStatedi (Joypad[J].Select) ?  32 : 0;
			PadState[1] |= !S9xGetStatedi (Joypad[J].Y)      ?  64 : 0;
			PadState[1] |= !S9xGetStatedi (Joypad[J].B)      ? 128 : 0;
			joypads [J] = PadState [0] | (PadState [1] << 8) | 0x80000000;
		}
		else
			joypads [J] = 0;
	}

	MapFuncPress = FALSE;
	for(i = 0; i < 8; i++) {
		if(!S9xGetStatedi(LOWORD(MapFunc[i]))) {
			if(!MapFuncPress) {
				MapFuncPress = TRUE;
				if(!MapFuncPressed) {
					MapFuncPressed = MapFuncPress;
					TurbofwdPressed[i] = (HIWORD(MapFunc[i]) == MF_TURBOPAUSEFWD);
					PostMessage (GUI.hWnd, WM_COMMAND, ID_MAPFUNC, HIWORD(MapFunc[i]));
					return;
				}
			}
		}
		else if(TurbofwdPressed[i] && HIWORD(MapFunc[i]) == MF_TURBOPAUSEFWD) {
			TurbofwdPressed[i] = FALSE;
			MapFuncPressed = MapFuncPress;
			PostMessage (GUI.hWnd, WM_COMMAND, ID_MAPFUNC, HIWORD(MapFunc[i]));
			return;
		}
	}
	MapFuncPressed = MapFuncPress;
}

static WORD GetdiJoyButtonCode()
{
	static WORD djb;
	static WORD dB;
	static bool dJoyButtonDownFlag;
	static bool dJoyPovDownFlag;
	static bool dJoyAxisxDownFlag;
	static bool dJoyAxisyDownFlag;
	static bool dJoyAxiszDownFlag;
	static bool dJoyAxisrDownFlag;
	static bool dJoyAxisuDownFlag;
	static bool dJoyAxisvDownFlag;
	static bool dJoyButtonDownedFlag;
	static bool dJoyPovDownedFlag;
	static bool dJoyAxisxDownedFlag;
	static bool dJoyAxisyDownedFlag;
	static bool dJoyAxiszDownedFlag;
	static bool dJoyAxisrDownedFlag;
	static bool dJoyAxisuDownedFlag;
	static bool dJoyAxisvDownedFlag;
	static unsigned long dJoyButtonCheck;
	HRESULT hr;

	djb = 0;
	dJoyButtonDownFlag = dJoyPovDownFlag = dJoyAxisxDownFlag = dJoyAxisyDownFlag = dJoyAxiszDownFlag = dJoyAxisrDownFlag = dJoyAxisuDownFlag = dJoyAxisvDownFlag = 0;
	++dJoyButtonCheck;
	for(WORD i = 0; i < diJoystick.num_joysticks; i++) {
		ClearJoyState(&diJoystick.joysticks[i].dijs);
		if(diJoystick.joysticks[i].did && diJoystick.joysticks[i].use_joystick && (dJoyButtonCheck % 8) == 0)
		{
			hr = diJoystick.joysticks[i].did->Poll();
			hr = diJoystick.joysticks[i].did->GetDeviceState(sizeof(DIJOYSTATE), &diJoystick.joysticks[i].dijs);

			if (FAILED(hr))
			{
				if (hr == DIERR_INPUTLOST || hr == DIERR_NOTACQUIRED)
				{
					hr = diJoystick.joysticks[i].did->Acquire();
				}
				continue;
			}

			for(dB = 0; dB < 32; dB++) {
				if((diJoystick.joysticks[i].dijs.rgbButtons[dB]  & 0x80) != 0)
				{
					dJoyButtonDownFlag = TRUE;
					if(dJoyButtonDownedFlag)
						break;
					djb = ((i << 8) | 0x8000 | (dB + 8));
					break;
				}
			}
			if(dJoyButtonDownFlag)
				break;
		}

		else if(diJoystick.joysticks[i].did && diJoystick.joysticks[i].use_joystick && (dJoyButtonCheck % 8) == 1 &&
				diJoystick.joysticks[i].num_pov)
		{
			hr = diJoystick.joysticks[i].did->Poll();
			hr = diJoystick.joysticks[i].did->GetDeviceState(sizeof(DIJOYSTATE), &diJoystick.joysticks[i].dijs);

			if (FAILED(hr))
			{
				if (hr == DIERR_INPUTLOST || hr == DIERR_NOTACQUIRED)
				{
					hr = diJoystick.joysticks[i].did->Acquire();
				}
				continue;
			}
			int pov_value;
			int angle;
			int axis_value_x, axis_value_y;
			for(WORD i2 = 0; i2 < diJoystick.joysticks[i].num_pov; i2++) {
				pov_value = (int)diJoystick.joysticks[i].dijs.rgdwPOV[i2];
				if(LOWORD(pov_value) != 0xffff) {
					angle = (pov_value + 27000) % 36000;
					angle = (36000 - angle) % 36000;
					angle /= 100;
					axis_value_x = 128 + (int)(127*cos(2*PI*angle/360.0)); // x
					axis_value_y = 128 + (int)(127*sin(2*PI*angle/360.0)); // y
					if(axis_value_x <= (128 - 128*DIDZ/100)) {
						dJoyPovDownFlag = TRUE;
						if(dJoyPovDownedFlag)
							break;
						djb = ((i << 8) | 0x8000 | 4);
						break;
					}
					if(axis_value_x >= (128 + 128*DIDZ/100)) {
						dJoyPovDownFlag = TRUE;
						if(dJoyPovDownedFlag)
							break;
						djb = ((i << 8) | 0x8000 | 5);
						break;
					}
					if(axis_value_y <= (128 - 128*DIDZ/100)) {
						dJoyPovDownFlag = TRUE;
						if(dJoyPovDownedFlag)
							break;
						djb = ((i << 8) | 0x8000 | 6);
						break;
					}
					if(axis_value_y >= (128 + 128*DIDZ/100)) {
						dJoyPovDownFlag = TRUE;
						if(dJoyPovDownedFlag)
							break;
						djb = ((i << 8) | 0x8000 | 7);
						break;
					}
				}
			}
			if(dJoyPovDownFlag)
				break;
		}
		else if(diJoystick.joysticks[i].did && diJoystick.joysticks[i].use_joystick && (dJoyButtonCheck % 8) == 2 &&
				diJoystick.joysticks[i].num_axes >= 1)
		{
			hr = diJoystick.joysticks[i].did->Poll();
			hr = diJoystick.joysticks[i].did->GetDeviceState(sizeof(DIJOYSTATE), &diJoystick.joysticks[i].dijs);

			if (FAILED(hr))
			{
				if (hr == DIERR_INPUTLOST || hr == DIERR_NOTACQUIRED)
				{
					hr = diJoystick.joysticks[i].did->Acquire();
				}
				continue;
			}
			//Left
			if(*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[0].offset) <= (128 - 128*DIDZ/100))
			{
				dJoyAxisxDownFlag = TRUE;
				if(dJoyAxisxDownedFlag)
					break;
				djb = ((i << 8) | 0x8000 | 0);
				break;
			}
			//Right
			else if(*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[0].offset) >= (128 + 128*DIDZ/100))
			{
				dJoyAxisxDownFlag = TRUE;
				if(dJoyAxisxDownedFlag)
					break;
				djb = ((i << 8) | 0x8000 | 1);
				break;
			}
		}
		else if(diJoystick.joysticks[i].did && diJoystick.joysticks[i].use_joystick && (dJoyButtonCheck % 8) == 3 &&
				diJoystick.joysticks[i].num_axes >= 2)
		{
			hr = diJoystick.joysticks[i].did->Poll();
			hr = diJoystick.joysticks[i].did->GetDeviceState(sizeof(DIJOYSTATE), &diJoystick.joysticks[i].dijs);

			if (FAILED(hr))
			{
				if (hr == DIERR_INPUTLOST || hr == DIERR_NOTACQUIRED)
				{
					hr = diJoystick.joysticks[i].did->Acquire();
				}
				continue;
			}
			//Up
			if(*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[1].offset) <= (128 - 128*DIDZ/100))
			{
				dJoyAxisyDownFlag = TRUE;
				if(dJoyAxisyDownedFlag)
					break;
				djb = ((i << 8) | 0x8000 | 2);
				break;
			}
			//Down
			else if(*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[1].offset) >= (128 + 128*DIDZ/100))
			{
				dJoyAxisyDownFlag = TRUE;
				if(dJoyAxisyDownedFlag)
					break;
				djb = ((i << 8) | 0x8000 | 3);
				break;
			}
		}
		else if(diJoystick.joysticks[i].did && diJoystick.joysticks[i].use_joystick && (dJoyButtonCheck % 8) == 4 &&
				diJoystick.joysticks[i].num_axes >= 3)
		{
			hr = diJoystick.joysticks[i].did->Poll();
			hr = diJoystick.joysticks[i].did->GetDeviceState(sizeof(DIJOYSTATE), &diJoystick.joysticks[i].dijs);

			if (FAILED(hr))
			{
				if (hr == DIERR_INPUTLOST || hr == DIERR_NOTACQUIRED)
				{
					hr = diJoystick.joysticks[i].did->Acquire();
				}
				continue;
			}
			//ZUp
			if(*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[2].offset) <= (128 - 128*DIDZ/100))
			{
				dJoyAxiszDownFlag = TRUE;
				if(dJoyAxiszDownedFlag)
					break;
				djb = ((i << 8) | 0x8000 | 41);
				break;
			}
			//ZDown
			else if(*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[2].offset) >= (128 + 128*DIDZ/100))
			{
				dJoyAxiszDownFlag = TRUE;
				if(dJoyAxiszDownedFlag)
					break;
				djb = ((i << 8) | 0x8000 | 42);
				break;
			}
		}
		else if(diJoystick.joysticks[i].did && diJoystick.joysticks[i].use_joystick && (dJoyButtonCheck % 8) == 5 &&
				diJoystick.joysticks[i].num_axes >= 4)
		{
			hr = diJoystick.joysticks[i].did->Poll();
			hr = diJoystick.joysticks[i].did->GetDeviceState(sizeof(DIJOYSTATE), &diJoystick.joysticks[i].dijs);

			if (FAILED(hr))
			{
				if (hr == DIERR_INPUTLOST || hr == DIERR_NOTACQUIRED)
				{
					hr = diJoystick.joysticks[i].did->Acquire();
				}
				continue;
			}
			//RUp
			if(*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[3].offset) <= (128 - 128*DIDZ/100))
			{
				dJoyAxisrDownFlag = TRUE;
				if(dJoyAxisrDownedFlag)
					break;
				djb = ((i << 8) | 0x8000 | 43);
				break;
			}
			//RDown
			else if(*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[3].offset) >= (128 + 128*DIDZ/100))
			{
				dJoyAxisrDownFlag = TRUE;
				if(dJoyAxisrDownedFlag)
					break;
				djb = ((i << 8) | 0x8000 | 44);
				break;
			}
		}
		else if(diJoystick.joysticks[i].did && diJoystick.joysticks[i].use_joystick && (dJoyButtonCheck % 8) == 6 &&
				diJoystick.joysticks[i].num_axes >= 5)
		{
			hr = diJoystick.joysticks[i].did->Poll();
			hr = diJoystick.joysticks[i].did->GetDeviceState(sizeof(DIJOYSTATE), &diJoystick.joysticks[i].dijs);

			if (FAILED(hr))
			{
				if (hr == DIERR_INPUTLOST || hr == DIERR_NOTACQUIRED)
				{
					hr = diJoystick.joysticks[i].did->Acquire();
				}
				continue;
			}
			//UUp
			if(*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[4].offset) <= (128 - 128*DIDZ/100))
			{
				dJoyAxisuDownFlag = TRUE;
				if(dJoyAxisuDownedFlag)
					break;
				djb = ((i << 8) | 0x8000 | 45);
				break;
			}
			//UDown
			else if(*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[4].offset) >= (128 + 128*DIDZ/100))
			{
				dJoyAxisuDownFlag = TRUE;
				if(dJoyAxisuDownedFlag)
					break;
				djb = ((i << 8) | 0x8000 | 46);
				break;
			}
		}
		else if(diJoystick.joysticks[i].did && diJoystick.joysticks[i].use_joystick && (dJoyButtonCheck % 8) == 7 &&
				diJoystick.joysticks[i].num_axes >= 6)
		{
			hr = diJoystick.joysticks[i].did->Poll();
			hr = diJoystick.joysticks[i].did->GetDeviceState(sizeof(DIJOYSTATE), &diJoystick.joysticks[i].dijs);

			if (FAILED(hr))
			{
				if (hr == DIERR_INPUTLOST || hr == DIERR_NOTACQUIRED)
				{
					hr = diJoystick.joysticks[i].did->Acquire();
				}
				continue;
			}
			//VUp
			if(*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[5].offset) <= (128 - 128*DIDZ/100))
			{
				dJoyAxisvDownFlag = TRUE;
				if(dJoyAxisvDownedFlag)
					break;
				djb = ((i << 8) | 0x8000 | 47);
				break;
			}
			//VDown
			else if(*(int *)(((byte *)&diJoystick.joysticks[i].dijs) + diJoystick.joysticks[i].axes[5].offset) >= (128 + 128*DIDZ/100))
			{
				dJoyAxisvDownFlag = TRUE;
				if(dJoyAxisvDownedFlag)
					break;
				djb = ((i << 8) | 0x8000 | 48);
				break;
			}
		}
	}
	if(dJoyButtonDownFlag)
		dJoyButtonDownedFlag = dJoyButtonDownFlag;
	else if(dJoyPovDownFlag)
		dJoyPovDownedFlag = dJoyPovDownFlag;
	else if(dJoyAxisxDownFlag)
		dJoyAxisxDownedFlag = dJoyAxisxDownFlag;
	else if(dJoyAxisyDownFlag)
		dJoyAxisyDownedFlag = dJoyAxisyDownFlag;
	else if(dJoyAxiszDownFlag)
		dJoyAxiszDownedFlag = dJoyAxiszDownFlag;
	else if(dJoyAxisrDownFlag)
		dJoyAxisrDownedFlag = dJoyAxisrDownFlag;
	else if(dJoyAxisuDownFlag)
		dJoyAxisuDownedFlag = dJoyAxisuDownFlag;
	else if(dJoyAxisvDownFlag)
		dJoyAxisvDownedFlag = dJoyAxisvDownFlag;
	if(!dJoyButtonDownFlag && (dJoyButtonCheck % 8) == 0)
		dJoyButtonDownedFlag = 0;
	else if(!dJoyPovDownFlag && (dJoyButtonCheck % 8) == 1)
		dJoyPovDownedFlag = 0;
	else if(!dJoyAxisxDownFlag && (dJoyButtonCheck % 8) == 2)
		dJoyAxisxDownedFlag = 0;
	else if(!dJoyAxisyDownFlag && (dJoyButtonCheck % 8) == 3)
		dJoyAxisyDownedFlag = 0;
	else if(!dJoyAxiszDownFlag && (dJoyButtonCheck % 8) == 4)
		dJoyAxiszDownedFlag = 0;
	else if(!dJoyAxisrDownFlag && (dJoyButtonCheck % 8) == 5)
		dJoyAxisrDownedFlag = 0;
	else if(!dJoyAxisuDownFlag && (dJoyButtonCheck % 8) == 6)
		dJoyAxisuDownedFlag = 0;
	else if(!dJoyAxisvDownFlag && (dJoyButtonCheck % 8) == 7)
		dJoyAxisvDownedFlag = 0;
	return djb;
}
