#include "../snes9x/snes9x.h"
#include "directx.h"
#include "render.h"
#include "../common/commutil.h"
#include "cfgdlg.h"

extern struct SSoundRates SoundRates[];
extern bool Draw1FrameFlag;

//proc
static void SettingsDlgCtrlSet(HWND hWnd);
static void Scale2X16RadioCtrlSet(HWND hWnd);
static void Scale2X8RadioCtrlSet(HWND hWnd);

//sound settings
static char GUI_SoundMode_cpy;
static int Settings_SoundBufferSize_cpy;
static int Settings_SoundMixInterval_cpy;

//Display settings
static int DoubleRenderingModes16_cpy;
static int DoubleRenderingModes8_cpy;
static unsigned char ForceWindowedDirectDrawClipper_cpy;
static unsigned char dRendering_cpy;
static unsigned char DirectX_DisableHardwareScale_cpy;
static bool GUI_DoubleBuffered_cpy;
static bool DirectX_OffScreenVRAM_cpy;
static bool DirectX_BackBufferVRAM_cpy;

//Thread
static DWORD GUI_PriorityClass_cpy;
static int GUI_ThreadPriority_cpy;

//Valuable extern
extern bool8 SupportHiResFlag_cpy;
extern bool8 SupportHiResFlag;
extern bool8 Settings_SixteenBit_cpy;
extern bool8 Options_GFXengine2_cpy;

extern bool8 DoubleRendering;
extern int DoubleRenderingModes16;
extern int DoubleRenderingModes8;
extern unsigned char dRendering;
extern unsigned char ForceWindowedDirectDrawClipper;
extern bool8 errConfigSwitch;
extern bool SPCPlayingFlag;
extern bool8 TransparencyFlag;
extern bool8 TransparencyFlag_cpy;

//proc extern
extern bool8 SetupSound( long rate, bool8 sixteen_bit, bool8 stereo);
extern bool8 S9xwSwitchScreenMode(bool8 Windowed);
extern void CurFullscreenDisplayModeIndexChange();
extern void S9xDrawScreenImage();

/*****************
 default settings
*****************/
extern char GUI_SoundMode_def;
extern int Settings_SoundBufferSize_def;
extern int Settings_SoundMixInterval_def;
extern int DoubleRenderingModes16_def;
extern int DoubleRenderingModes8_def;
extern unsigned char ForceWindowedDirectDrawClipper_def;
extern unsigned char dRendering_def;
extern bool GUI_DoubleBuffered_def;
extern DWORD GUI_PriorityClass_def;
extern int GUI_ThreadPriority_def;

void S9xwSettingsLoad()
{
	//sound settings
	GUI_SoundMode_cpy = GUI.SoundMode;
	Settings_SoundBufferSize_cpy = Settings.SoundBufferSize;
	Settings_SoundMixInterval_cpy = Settings.SoundMixInterval;

	//Display settings
	DoubleRenderingModes16_cpy = DoubleRenderingModes16;
	DoubleRenderingModes8_cpy = DoubleRenderingModes8;
	ForceWindowedDirectDrawClipper_cpy = ForceWindowedDirectDrawClipper;
	dRendering_cpy = dRendering;
	DirectX_DisableHardwareScale_cpy = DirectX.DisableHardwareScale;
	GUI_DoubleBuffered_cpy = GUI.DoubleBuffered;
	DirectX_OffScreenVRAM_cpy = DirectX.OffScreenVRAM;
	DirectX_BackBufferVRAM_cpy = DirectX.BackBufferVRAM;

	//Thread
	GUI_PriorityClass_cpy = GUI.PriorityClass;
	GUI_ThreadPriority_cpy = GUI.ThreadPriority;
}

static HWND hSettingsDlg;

void S9xwDefaultSettingsLoad()
{
	//sound settings
	GUI_SoundMode_cpy = GUI_SoundMode_def;
	Settings_SoundBufferSize_cpy = Settings_SoundBufferSize_def;
	Settings_SoundMixInterval_cpy = Settings_SoundMixInterval_def;

	//Display settings
	DoubleRenderingModes16_cpy = DoubleRenderingModes16_def;
	DoubleRenderingModes8_cpy = DoubleRenderingModes8_def;
	ForceWindowedDirectDrawClipper_cpy = ForceWindowedDirectDrawClipper_def;
	dRendering_cpy = dRendering_def;
	DirectX_DisableHardwareScale_cpy = DirectX_DisableHardwareScale_def;
	GUI_DoubleBuffered_cpy = GUI_DoubleBuffered_def;
	DirectX_OffScreenVRAM_cpy = DirectX_OffScreenVRAM_def;
	DirectX_BackBufferVRAM_cpy = DirectX_BackBufferVRAM_def;

	//Thread
	GUI_PriorityClass_cpy = GUI_PriorityClass_def;
	GUI_ThreadPriority_cpy = GUI_ThreadPriority_def;

	SettingsDlgCtrlSet(hSettingsDlg);
}

void S9xwSettingsSave()
{
	SelectRenderingCodes();
	ForceWindowedDirectDrawClipper = ForceWindowedDirectDrawClipper_cpy;
	dRendering = dRendering_cpy;
	if(dRendering)
		DirectX.ForceWindowedDirectDrawClipper = FALSE;
	else
		DirectX.ForceWindowedDirectDrawClipper = ForceWindowedDirectDrawClipper;
	DirectX.DisableHardwareScale = DirectX_DisableHardwareScale_cpy;
	GUI.DoubleBuffered = GUI_DoubleBuffered_cpy;
	DirectX.BackBufferVRAM = DirectX_BackBufferVRAM_cpy;
	TransparencyFlag = TransparencyFlag_cpy;
	if (Settings_SixteenBit_cpy)
		Settings.Transparency = TransparencyFlag;
	else
		Settings.Transparency = FALSE;
	if(Settings.SixteenBit != Settings_SixteenBit_cpy)
		Draw1FrameFlag = FALSE;
	if(DoubleRenderingModes16_cpy != DoubleRenderingModes16 || DoubleRenderingModes8_cpy != DoubleRenderingModes8 || SupportHiResFlag_cpy != SupportHiResFlag
		|| Settings.SixteenBit != Settings_SixteenBit_cpy || DirectX.OffScreenVRAM != DirectX_OffScreenVRAM_cpy) {
		DoubleRenderingModes16 = DoubleRenderingModes16_cpy;
		DoubleRenderingModes8 = DoubleRenderingModes8_cpy;
		DirectX.OffScreenVRAM = DirectX_OffScreenVRAM_cpy;
		SupportHiResFlag = SupportHiResFlag_cpy;
		if(Options.GFXengine2 != Options_GFXengine2_cpy || Settings.SixteenBit != Settings_SixteenBit_cpy) {
			Settings.SixteenBit = Settings_SixteenBit_cpy;
			Options.GFXengine2 = Options_GFXengine2_cpy;
			S9xSwitchGFXengine(Options.GFXengine2);
		}
		if(!DoubleRendering) {
			GUI.Scale = 0;
			Settings.SupportHiRes = FALSE;
		}
		else {
			//2xSaI[hւ̐؂芷̏ꍇ̃XN[obt@NA΍
			ResetRenderMethod();
			GUI.Scale = (Settings.SixteenBit ? DoubleRenderingModes16 : DoubleRenderingModes8);
			Settings.SupportHiRes = SupportHiResFlag;
		}
		CurFullscreenDisplayModeIndexChange();
		if(GUI.FullScreen)
		{
			if(!(errConfigSwitch = !S9xwSwitchScreenMode(TRUE)))
				GUI.FullScreen = TRUE;
		}
		else
		{
			if(!(errConfigSwitch = !S9xwSwitchScreenMode(TRUE)))
				GUI.FullScreen = FALSE;
		}
		if(!errConfigSwitch && (GUI.StopEmulation || SPCPlayingFlag))
			S9xDrawScreenImage();
	}

	if(Options.GFXengine2 != Options_GFXengine2_cpy) {
		Options.GFXengine2 = Options_GFXengine2_cpy;
		S9xSwitchGFXengine(Options.GFXengine2);
	}

	if(GUI.SoundMode != GUI_SoundMode_cpy || Settings.SoundBufferSize != Settings_SoundBufferSize_cpy || Settings.SoundMixInterval != Settings_SoundMixInterval_cpy) {
		GUI.SoundMode = GUI_SoundMode_cpy;
		Settings.SoundBufferSize = Settings_SoundBufferSize_cpy;
		Settings.SoundMixInterval = Settings_SoundMixInterval_cpy;
		SetupSound( SoundRates[GUI.SoundMode].rate, SoundRates[GUI.SoundMode].sixteen_bit, SoundRates[GUI.SoundMode].stereo);
	}
	GUI.PriorityClass = GUI_PriorityClass_cpy;
	GUI.ThreadPriority = GUI_ThreadPriority_cpy;
}


LRESULT CALLBACK SettingsDlgProc(HWND hWnd, UINT uMsg,WPARAM wParam, LPARAM lParam)
{
	static int i;
	static char *SoundModeStr[] = {"8000Hz 8bit mono", "11025Hz 8bit mono", "22050Hz 8bit mono", "32000Hz 8bit mono", "44100Hz 8bit mono", "48000Hz 8bit mono",
									 "8000Hz 16bit mono", "11025Hz 16bit mono", "22050Hz 16bit mono", "32000Hz 16bit mono", "44100Hz 16bit mono", "48000Hz 16bit mono",
									"8000Hz 8bit stereo", "11025Hz 8bit stereo", "22050Hz 8bit stereo", "32000Hz 8bit stereo", "44100Hz 8bit stereo", "48000Hz 8bit stereo",
									"8000Hz 16bit stereo", "11025Hz 16bit stereo", "22050Hz 16bit stereo", "32000Hz 16bit stereo", "44100Hz 16bit stereo", "48000Hz 16bit stereo"};
	static char *SoundBufferLengthStr[] = {"10 ms", "20 ms", "40 ms", "80 ms", "160 ms", "320 ms", "640 ms"};
	static char *SoundMixIntervalStr[] = {"10 ms", "20 ms", "30 ms", "40 ms", "50 ms", "60 ms", "70 ms", "80 ms", "90 ms", "100 ms", "110 ms", "120 ms", "130 ms", "140 ms", "150 ms"};
	static char *PriorityClassStr[] = {"IDLE_PRIORITY_CLASS", "NORMAL_PRIORITY_CLASS", "HIGH_PRIORITY_CLASS", "REALTIME_PRIORITY_CLASS"};
	static char *ThreadPriorityStr[] = {"THREAD_PRIORITY_LOWEST", "THREAD_PRIORITY_BELOW_NORMAL", "THREAD_PRIORITY_NORMAL", "THREAD_PRIORITY_ABOVE_NORMAL", "THREAD_PRIORITY_HIGHEST", "THREAD_PRIORITY_IDLE", "THREAD_PRIORITY_TIME_CRITICAL"};

	switch (uMsg) {
		case WM_INITDIALOG:
			hSettingsDlg = hWnd;
			//TEh[hR{{bNXɃeLXgZbg
			for ( i = 0; i < 24; i++)
				SendMessage(GetDlgItem(hWnd, IDC_COMBO_SOUNDMODE), CB_INSERTSTRING, (WPARAM)i, (LPARAM)SoundModeStr[i]);
			//SoundBufferLengthR{{bNXɃeLXgZbg
			for ( i = 0; i < 7; i++)
				SendMessage(GetDlgItem(hWnd, IDC_COMBO_SOUNDBUFFERLENGTH), CB_INSERTSTRING, (WPARAM)i, (LPARAM)SoundBufferLengthStr[i]);
			//SoundMixIntervalR{{bNXɃeLXgZbg
			for ( i = 0; i < 15; i++)
				SendMessage(GetDlgItem(hWnd, IDC_COMBO_SOUNDMIXINTERVAL), CB_INSERTSTRING, (WPARAM)i, (LPARAM)SoundMixIntervalStr[i]);

			//Thread
			for (i = 0; i < 4; i++)
				SendMessage(GetDlgItem(hWnd, IDC_COMBO_PRIORITYCLASS), CB_INSERTSTRING, (WPARAM)i, (LPARAM)PriorityClassStr[i]);
			for (i = 0; i < 7; i++)
				SendMessage(GetDlgItem(hWnd, IDC_COMBO_THREADPRIORITY), CB_INSERTSTRING, (WPARAM)i, (LPARAM)ThreadPriorityStr[i]);
			
			SettingsDlgCtrlSet(hWnd);
			break;

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDC_RADIO_2X_HARDWARE:
				case IDC_RADIO_2X_SOFTWARE:
				case IDC_RADIO_2X_SCANLINES:
				case IDC_RADIO_2X_EAGLE:
				case IDC_RADIO_2X_SAI:
				case IDC_RADIO_SUPER_2X_SAI:
				case IDC_RADIO_SCALE_2X_SAI:
				case IDC_RADIO_2X_TVMODE:
				case IDC_RADIO_2X_HQ2X:
				case IDC_RADIO_2X_LQ2X:

				case IDC_CHECK_RENDERING_ASMMMX:
				case IDC_CHECK_FORCEWINDOWEDDIRECTDRAWCLIPPER:
				case IDC_CHECK_DOUBLEBUFFERING:
				case IDC_CHECK_DIRECTSCREENRENDERING:
				case IDC_CHECK_DISABLEHARDWARESCALE:
				case IDC_CHECK_OFFSCREENVRAM:
				case IDC_CHECK_BACKBUFFERVRAM:
					if(HIWORD(wParam) == BN_CLICKED) {
						bool8 tmpVal = (bool8)(IsDlgButtonChecked(hWnd, (int)LOWORD(wParam)) == BST_CHECKED);
						if(LOWORD(wParam) == IDC_RADIO_2X_HARDWARE ||
								LOWORD(wParam) == IDC_RADIO_2X_SOFTWARE ||
								LOWORD(wParam) == IDC_RADIO_2X_SCANLINES ||
								LOWORD(wParam) == IDC_RADIO_2X_EAGLE ||
								LOWORD(wParam) == IDC_RADIO_2X_SAI ||
								LOWORD(wParam) == IDC_RADIO_SUPER_2X_SAI ||
								LOWORD(wParam) == IDC_RADIO_SCALE_2X_SAI ||
								LOWORD(wParam) == IDC_RADIO_2X_TVMODE ||
								LOWORD(wParam) == IDC_RADIO_2X_HQ2X ||
								LOWORD(wParam) == IDC_RADIO_2X_LQ2X) {
							if(LOWORD(wParam) == IDC_RADIO_2X_HARDWARE) {
								if(Settings_SixteenBit_cpy)
									DoubleRenderingModes16_cpy = 1;
								else
									DoubleRenderingModes8_cpy = 1;
							}
							else if(LOWORD(wParam) == IDC_RADIO_2X_SOFTWARE) {
								if(Settings_SixteenBit_cpy)
									DoubleRenderingModes16_cpy = 2;
								else
									DoubleRenderingModes8_cpy = 2;
							}
							else if(LOWORD(wParam) == IDC_RADIO_2X_SCANLINES) {
								if(Settings_SixteenBit_cpy)
									DoubleRenderingModes16_cpy = 3;
								else
									DoubleRenderingModes8_cpy = 3;
							}
							else if(LOWORD(wParam) == IDC_RADIO_2X_EAGLE)
								DoubleRenderingModes16_cpy = 4;
							else if(LOWORD(wParam) == IDC_RADIO_2X_SAI)
								DoubleRenderingModes16_cpy = 5;
							else if(LOWORD(wParam) == IDC_RADIO_SUPER_2X_SAI)
								DoubleRenderingModes16_cpy = 6;
							else if(LOWORD(wParam) == IDC_RADIO_SCALE_2X_SAI)
								DoubleRenderingModes16_cpy = 7;
							else if(LOWORD(wParam) == IDC_RADIO_2X_TVMODE)
								DoubleRenderingModes16_cpy = 8;
							else if(LOWORD(wParam) == IDC_RADIO_2X_HQ2X)
								DoubleRenderingModes16_cpy = 9;
							else if(LOWORD(wParam) == IDC_RADIO_2X_LQ2X)
								DoubleRenderingModes16_cpy = 10;
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_HARDWARE), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_HARDWARE ? 1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_SOFTWARE), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_SOFTWARE ? 1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_SCANLINES), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_SCANLINES ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_EAGLE), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_EAGLE ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_SAI), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_SAI ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_SUPER_2X_SAI), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_SUPER_2X_SAI ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_SCALE_2X_SAI), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_SCALE_2X_SAI ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_TVMODE), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_TVMODE ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_HQ2X), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_HQ2X ?  1 : 0, 0);
							SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_LQ2X), BM_SETCHECK, LOWORD(wParam) == IDC_RADIO_2X_LQ2X ?  1 : 0, 0);
							EnableWindow(GetDlgItem(hWnd, IDC_CHECK_DISABLEHARDWARESCALE), (Settings_SixteenBit_cpy ? (DoubleRenderingModes16_cpy == 1) : (DoubleRenderingModes8_cpy == 1)) ? FALSE : TRUE);
							if(DoubleRenderingModes16_cpy == DoubleRenderingModes16 && DoubleRenderingModes8_cpy == DoubleRenderingModes8)
								break;
						}
						else if(LOWORD(wParam) == IDC_CHECK_FORCEWINDOWEDDIRECTDRAWCLIPPER)
						{
							ForceWindowedDirectDrawClipper_cpy = tmpVal;
						}
						else if(LOWORD(wParam) == IDC_CHECK_DOUBLEBUFFERING)
						{
							GUI_DoubleBuffered_cpy = (tmpVal ? TRUE : FALSE);
						}
						else if(LOWORD(wParam) == IDC_CHECK_DIRECTSCREENRENDERING)
						{
							dRendering_cpy = tmpVal;
							EnableWindow(GetDlgItem(hWnd, IDC_CHECK_FORCEWINDOWEDDIRECTDRAWCLIPPER), dRendering_cpy ? FALSE : TRUE);
						}
						else if(LOWORD(wParam) == IDC_CHECK_DISABLEHARDWARESCALE)
						{
							DirectX_DisableHardwareScale_cpy = (unsigned char)tmpVal;
						}
						else if(LOWORD(wParam) == IDC_CHECK_OFFSCREENVRAM)
						{
							DirectX_OffScreenVRAM_cpy = (tmpVal ? TRUE : FALSE);
						}
						else if(LOWORD(wParam) == IDC_CHECK_BACKBUFFERVRAM)
						{
							DirectX_BackBufferVRAM_cpy = (tmpVal ? TRUE : FALSE);
						}
						ConfigDlgSettingsChanged(hWnd);
					}
					break;

				case IDC_COMBO_SOUNDMODE:
				case IDC_COMBO_SOUNDBUFFERLENGTH:
				case IDC_COMBO_SOUNDMIXINTERVAL:
				case IDC_COMBO_PRIORITYCLASS:
				case IDC_COMBO_THREADPRIORITY:
					if(HIWORD(wParam) == CBN_SELENDOK) {
						i = (int)SendMessage(GetDlgItem(hWnd, (int)LOWORD(wParam)), CB_GETCURSEL, 0, 0);
						if(LOWORD(wParam) == IDC_COMBO_SOUNDMODE)
							GUI_SoundMode_cpy = (uint8)i;
						else if(LOWORD(wParam) == IDC_COMBO_SOUNDBUFFERLENGTH)
							Settings_SoundBufferSize_cpy = (1 << i);
						else if(LOWORD(wParam) == IDC_COMBO_SOUNDMIXINTERVAL)
							Settings_SoundMixInterval_cpy = (i + 1) * 10;
						else if(LOWORD(wParam) == IDC_COMBO_PRIORITYCLASS) {
							if(i == 0)
								GUI_PriorityClass_cpy = IDLE_PRIORITY_CLASS;
							else if(i == 1)
								GUI_PriorityClass_cpy = NORMAL_PRIORITY_CLASS;
							else if(i == 2)
								GUI_PriorityClass_cpy = HIGH_PRIORITY_CLASS;
							else if(i == 3)
								GUI_PriorityClass_cpy = REALTIME_PRIORITY_CLASS;
						}
						else if(LOWORD(wParam) == IDC_COMBO_THREADPRIORITY) {
							if(i == 0)
								GUI_ThreadPriority_cpy = THREAD_PRIORITY_LOWEST;
							else if(i == 1)
								GUI_ThreadPriority_cpy = THREAD_PRIORITY_BELOW_NORMAL;
							else if(i == 2)
								GUI_ThreadPriority_cpy = THREAD_PRIORITY_NORMAL;
							else if(i == 3)
								GUI_ThreadPriority_cpy = THREAD_PRIORITY_ABOVE_NORMAL;
							else if(i == 4)
								GUI_ThreadPriority_cpy = THREAD_PRIORITY_HIGHEST;
							else if(i == 5)
								GUI_ThreadPriority_cpy = THREAD_PRIORITY_IDLE;
							else if(i == 6)
								GUI_ThreadPriority_cpy = THREAD_PRIORITY_TIME_CRITICAL;
						}
						if(GUI_SoundMode_cpy != GUI.SoundMode || Settings_SoundBufferSize_cpy != Settings.SoundBufferSize ||
							Settings_SoundMixInterval_cpy != Settings.SoundMixInterval || GUI_PriorityClass_cpy != GUI.PriorityClass ||
							GUI_ThreadPriority_cpy != GUI.ThreadPriority) {
							ConfigDlgSettingsChanged(hWnd);
						}
					}
					break;

				default:
					break;
			}
			break;

		case WM_NOTIFY:
			switch(((NMHDR *)lParam)->code) {
				case PSN_APPLY:
					ConfigDlgSettingsSave();
					break;

				case PSN_HELP:
					ConfigDlgDefaultSettingsLoad(hWnd);
					break;

				case PSN_QUERYCANCEL:
					break;

				case PSN_SETACTIVE:
					if(Settings_SixteenBit_cpy)
						Scale2X16RadioCtrlSet(hWnd);
					else
						Scale2X8RadioCtrlSet(hWnd);
					EnableWindow(GetDlgItem(hWnd, IDC_CHECK_DISABLEHARDWARESCALE), (Settings_SixteenBit_cpy ? (DoubleRenderingModes16_cpy == 1) : (DoubleRenderingModes8_cpy == 1)) ? FALSE : TRUE);
					StartPageIndex = 0;
					break;

				default:
					break;
			}
			break;

		default:
			break;
	}
	return FALSE;
}

static void SettingsDlgCtrlSet(HWND hWnd)
{
	int i, c;
	SendMessage(GetDlgItem(hWnd, IDC_COMBO_SOUNDMODE), CB_SETCURSEL, (WPARAM)GUI_SoundMode_cpy, 0);
	for(c = i = 1; i != Settings_SoundBufferSize_cpy && i < 64; c++, i <<= 1);
	SendMessage(GetDlgItem(hWnd, IDC_COMBO_SOUNDBUFFERLENGTH), CB_SETCURSEL, (WPARAM)(c - 1), 0);
	i = ((Settings_SoundMixInterval_cpy / 10) - 1);
	SendMessage(GetDlgItem(hWnd, IDC_COMBO_SOUNDMIXINTERVAL), CB_SETCURSEL, (WPARAM)i, 0);

	if(Settings_SixteenBit_cpy)
		Scale2X16RadioCtrlSet(hWnd);
	else
		Scale2X8RadioCtrlSet(hWnd);

	SendMessage(GetDlgItem(hWnd, IDC_CHECK_FORCEWINDOWEDDIRECTDRAWCLIPPER), BM_SETCHECK, (WPARAM)(ForceWindowedDirectDrawClipper_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_DOUBLEBUFFERING), BM_SETCHECK, (WPARAM)(GUI_DoubleBuffered_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_DIRECTSCREENRENDERING), BM_SETCHECK, (WPARAM)(dRendering_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_DISABLEHARDWARESCALE), BM_SETCHECK, (WPARAM)(DirectX_DisableHardwareScale_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_OFFSCREENVRAM), BM_SETCHECK, (WPARAM)(DirectX_OffScreenVRAM_cpy ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_CHECK_BACKBUFFERVRAM), BM_SETCHECK, (WPARAM)(DirectX_BackBufferVRAM_cpy ? 1 : 0), 0);

	EnableWindow(GetDlgItem(hWnd, IDC_CHECK_FORCEWINDOWEDDIRECTDRAWCLIPPER), dRendering_cpy ? FALSE : TRUE);
	EnableWindow(GetDlgItem(hWnd, IDC_CHECK_DISABLEHARDWARESCALE), (Settings_SixteenBit_cpy ? (DoubleRenderingModes16_cpy == 1) : (DoubleRenderingModes8_cpy == 1)) ? FALSE : TRUE);

	if(GUI_PriorityClass_cpy == IDLE_PRIORITY_CLASS)
		i = 0;
	else if(GUI_PriorityClass_cpy == NORMAL_PRIORITY_CLASS)
		i = 1;
	else if(GUI_PriorityClass_cpy == HIGH_PRIORITY_CLASS)
		i = 2;
	else if(GUI_PriorityClass_cpy == REALTIME_PRIORITY_CLASS)
		i = 3;
	else
		i = 1;
	SendMessage(GetDlgItem(hWnd, IDC_COMBO_PRIORITYCLASS), CB_SETCURSEL, (WPARAM)i, 0);

	if(GUI_ThreadPriority_cpy == THREAD_PRIORITY_LOWEST)
		i = 0;
	else if(GUI_ThreadPriority_cpy == THREAD_PRIORITY_BELOW_NORMAL)
		i = 1;
	else if(GUI_ThreadPriority_cpy == THREAD_PRIORITY_NORMAL)
		i = 2;
	else if(GUI_ThreadPriority_cpy == THREAD_PRIORITY_ABOVE_NORMAL)
		i = 3;
	else if(GUI_ThreadPriority_cpy == THREAD_PRIORITY_HIGHEST)
		i = 4;
	else if(GUI_ThreadPriority_cpy == THREAD_PRIORITY_IDLE)
		i = 5;
	else if(GUI_ThreadPriority_cpy == THREAD_PRIORITY_TIME_CRITICAL)
		i = 6;
	else
		i = 2;
	SendMessage(GetDlgItem(hWnd, IDC_COMBO_THREADPRIORITY), CB_SETCURSEL, (WPARAM)i, 0);
}

static void Scale2X16RadioCtrlSet(HWND hWnd)
{
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_HARDWARE), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 1) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_SOFTWARE), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 2) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_SCANLINES), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 3) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_EAGLE), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 4) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_SAI), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 5) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_SUPER_2X_SAI), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 6) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_SCALE_2X_SAI), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 7) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_TVMODE), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 8) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_HQ2X), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 9) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_LQ2X), BM_SETCHECK, (WPARAM)((DoubleRenderingModes16_cpy == 10) ? 1 : 0), 0);
	EnableWindow(GetDlgItem(hWnd, IDC_RADIO_2X_EAGLE), TRUE);
	EnableWindow(GetDlgItem(hWnd, IDC_RADIO_2X_SAI), TRUE);
	EnableWindow(GetDlgItem(hWnd, IDC_RADIO_SUPER_2X_SAI), TRUE);
	EnableWindow(GetDlgItem(hWnd, IDC_RADIO_SCALE_2X_SAI), TRUE);
	EnableWindow(GetDlgItem(hWnd, IDC_RADIO_2X_TVMODE), TRUE);
	EnableWindow(GetDlgItem(hWnd, IDC_RADIO_2X_HQ2X), TRUE);
	EnableWindow(GetDlgItem(hWnd, IDC_RADIO_2X_LQ2X), TRUE);
}

static void Scale2X8RadioCtrlSet(HWND hWnd)
{
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_HARDWARE), BM_SETCHECK, (WPARAM)((DoubleRenderingModes8_cpy == 1) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_SOFTWARE), BM_SETCHECK, (WPARAM)((DoubleRenderingModes8_cpy == 2) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_SCANLINES), BM_SETCHECK, (WPARAM)((DoubleRenderingModes8_cpy == 3) ? 1 : 0), 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_EAGLE), BM_SETCHECK, 0, 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_SAI), BM_SETCHECK, 0, 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_SUPER_2X_SAI), BM_SETCHECK, 0, 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_SCALE_2X_SAI), BM_SETCHECK, 0, 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_TVMODE), BM_SETCHECK, 0, 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_HQ2X), BM_SETCHECK, 0, 0);
	SendMessage(GetDlgItem(hWnd, IDC_RADIO_2X_LQ2X), BM_SETCHECK, 0, 0);
	EnableWindow(GetDlgItem(hWnd, IDC_RADIO_2X_EAGLE), FALSE);
	EnableWindow(GetDlgItem(hWnd, IDC_RADIO_2X_SAI), FALSE);
	EnableWindow(GetDlgItem(hWnd, IDC_RADIO_SUPER_2X_SAI), FALSE);
	EnableWindow(GetDlgItem(hWnd, IDC_RADIO_SCALE_2X_SAI), FALSE);
	EnableWindow(GetDlgItem(hWnd, IDC_RADIO_2X_TVMODE), FALSE);
	EnableWindow(GetDlgItem(hWnd, IDC_RADIO_2X_HQ2X), FALSE);
	EnableWindow(GetDlgItem(hWnd, IDC_RADIO_2X_LQ2X), FALSE);
}
