#include "../snes9x/snes9x.h"
#include "directx.h"
#include "joywin.h"
#include "render.h"
#include "../snes9x/memmap.h"
#include "../snes9x/debug.h"
#include "../snes9x/cpuexec.h"
#include "../snes9x/ppu.h"
#include "../snes9x/snapshot.h"
#include "../snes9x/apu.h"
#include "../snes9x/display.h"
#include "../snes9x/gfx.h"
#include "../snes9x/soundux.h"
#include "../snes9x/netplay.h"

#ifdef FMOD_SUPPORT
#include "fmod.h"
#endif

bool8 RealSetupSound (long rate, bool8 sixteen_bit, bool8 stereo);
void S9xSetWinPixelFormat ();

typedef struct
{
    uint8 red;
    uint8 green;
    uint8 blue;
} Colour;

static Colour FixedColours [256];
extern "C" {
uint8 S9xwPalette [0x10000];
int color_diff [0x10000];
};

FILE *trace_fs = NULL;

void S9xTextMode( void)
{
}

void S9xGraphicsMode ()
{
}

void S9xExit( void)
{
	MessageBox(NULL, Languages[ GUI.Language].S9xExitMessage, "Snes9XW", MB_OK | MB_ICONSTOP);
	SendMessage( GUI.hWnd, WM_COMMAND, ID_FILE_EXIT, 0);
}

extern "C" const char *S9xGetFilename (const char *e)
{
    static char filename [_MAX_PATH + 1];
    char drive [_MAX_DRIVE + 1];
    char dir [_MAX_DIR + 1];
    char fname [_MAX_FNAME + 1];
    char ext [_MAX_EXT + 1];

    if (strlen (GUI.FreezeFileDir))
    {
        _splitpath (Memory.ROMFilename, drive, dir, fname, ext);
        strcpy (filename, GUI.FreezeFileDir);
        strcat (filename, "\\");
        strcat (filename, fname);
        strcat (filename, e);
    }
    else
    {
        _splitpath (Memory.ROMFilename, drive, dir, fname, ext);
        _makepath( filename, drive, dir, fname, e);
    }

    return (filename);
}

char SPCDumpDirectory[_MAX_PATH] = {0};

static int spcnum = 0;

char *S9xwGetSPCFilename ()
{
    static char filename [_MAX_PATH + 1];
    char drive [_MAX_DRIVE + 1];
    char dir [_MAX_DIR + 1];
    char fname [_MAX_FNAME + 1];
    char ext [_MAX_EXT + 1];


    if (strlen (SPCDumpDirectory))
    {
        _splitpath (Memory.ROMFilename, drive, dir, fname, ext);
        strcpy (filename, SPCDumpDirectory);
        strcat (filename, "\\");
        strcat (filename, fname);
		strcat (filename, "_000");
        strcat (filename, ".spc");
		_splitpath (filename, drive, dir, fname, ext);
    }
    else
    {
        _splitpath (Memory.ROMFilename, drive, dir, fname, ext);
		strcat (fname, "_000");
        _makepath( filename, drive, dir, fname, ".spc");
		_splitpath (filename, drive, dir, fname, ext);
    }
	spcnum = 0;
	while((int)GetFileAttributes(filename) != -1 && spcnum != 999)
	{
		++spcnum;
		wsprintf(&fname[lstrlen(fname) - 3], "%03d", spcnum);
		_makepath( filename, drive, dir, fname, ext);
	}
    return (filename);
}


static char SavedSPCNumberAndExtStr[10];

char *S9xwGetSavedSPCNumberAndExtStr()
{
	SavedSPCNumberAndExtStr[0] = '_';
	wsprintf(&SavedSPCNumberAndExtStr[1], "%03d", spcnum);
	lstrcat(SavedSPCNumberAndExtStr, ".spc");
	return SavedSPCNumberAndExtStr;
}

char ScreenShotDirectory[_MAX_PATH] = {0};

static char imgext [_MAX_EXT] = ".bmp";
static int imgnum = 0;

char *S9xwGetScreenImageFilename (char *ImageTypeExt)
{
	static char filename [_MAX_PATH + 1];
	char drive [_MAX_DRIVE + 1];
	char dir [_MAX_DIR + 1];
	char fname [_MAX_FNAME + 1];

	_splitpath (Memory.ROMFilename, drive, dir, fname, imgext);
	lstrcpy(imgext, ImageTypeExt);
	CharLower(imgext);

	if (strlen (ScreenShotDirectory))
	{
		strcpy (filename, ScreenShotDirectory);
		strcat (filename, "\\");
		strcat (filename, fname);
		strcat (filename, "_000");
		strcat (filename, imgext);
		_splitpath (filename, drive, dir, fname, imgext);
	}
	else
	{
		strcat (fname, "_000");
		_makepath( filename, drive, dir, fname, imgext);
		_splitpath (filename, drive, dir, fname, imgext);
	}
	
	imgnum = 0;
	while((int)GetFileAttributes(filename) != -1 && imgnum != 999)
	{
		++imgnum;
		wsprintf(&fname[lstrlen(fname) - 3], "%03d", imgnum);
		_makepath( filename, drive, dir, fname, imgext);
	}

	return (filename);
}

static char SavedScreenImageNumberAndTypeStr[10];

char *S9xwGetSavedScreenImageNumberAndTypeStr()
{
	SavedScreenImageNumberAndTypeStr[0] = '_';
	wsprintf(&SavedScreenImageNumberAndTypeStr[1], "%03d", imgnum);
	lstrcat(SavedScreenImageNumberAndTypeStr, imgext);
	return SavedScreenImageNumberAndTypeStr;
}

bool8 S9xOpenSnapshotFile( const char *fname, bool8 read_only, STREAM *file)
{
    char filename [_MAX_PATH + 1];
    char drive [_MAX_DRIVE + 1];
    char dir [_MAX_DIR + 1];
    char fn [_MAX_FNAME + 1];
    char ext [_MAX_EXT + 1];

    _splitpath( fname, drive, dir, fn, ext);
    _makepath( filename, drive, dir, fn, ext[0] == '\0' ? ".000" : ext);

    if (read_only)
    {
	if ((*file = OPEN_STREAM (filename, "rb")))
	    return (TRUE);
    }
    else
    {
	if ((*file = OPEN_STREAM (filename, "wb")))
	    return (TRUE);
        FILE *fs = fopen (filename, "rb");
        if (fs)
        {
            sprintf (String, "Freeze file \"%s\" exists but is read only",
                     filename);
            fclose (fs);
            S9xMessage (S9X_ERROR, S9X_FREEZE_FILE_NOT_FOUND, String);
        }
        else
        {
            sprintf (String, "Cannot create freeze file \"%s\". Directory is read-only or does not exist.", filename);
            
            S9xMessage (S9X_ERROR, S9X_FREEZE_FILE_NOT_FOUND, String);
        }
    }
    return (FALSE);
}

void S9xCloseSnapshotFile( STREAM file)
{
    CLOSE_STREAM (file);
}

void S9xMessage (int, int, const char *str)
{
#ifdef DEBUGGER
    static FILE *out = NULL;

    if (out == NULL)
        out = fopen ("out.txt", "w");

    fprintf (out, "%s\n", str);
#endif

    S9xSetInfoString (str);
}

unsigned long _interval = 10;
long _buffernos = 4;
long _blocksize = 4400;
long _samplecount = 440;
long _maxsamplecount = 0;
long _buffersize = 0;
uint8 *SoundBuffer = NULL;
bool StartPlaying = false;
DWORD _lastblock = 0;
bool8 pending_setup = false;
long pending_rate = 0;
bool8 pending_16bit = false;
bool8 pending_stereo = false;

bool8 SetupSound (long rate, bool8 sixteen_bit, bool8 stereo)
{
#ifdef FMOD_SUPPORT
    if (Settings.SoundDriver >= WIN_FMOD_DIRECT_SOUND_DRIVER)
    {
        so.mute_sound = TRUE;
        if (SoundBuffer)
        {
            SoundBuffer -= 32 * 1024;
            free (SoundBuffer);
            SoundBuffer = NULL;
        }
        
        _interval = 20;
        if (Settings.SoundBufferSize < 1)
            Settings.SoundBufferSize = 1;
        if (Settings.SoundBufferSize > 64)
            Settings.SoundBufferSize = 64;
        
        _buffernos = 4 * Settings.SoundBufferSize;
        int s = (rate * _interval *
                 (Settings.Stereo ? 2 : 1) * 
                 (Settings.SixteenBitSound ? 2 : 1)) / 1000;
        
        _blocksize = 64;
        while (_blocksize < s)
            _blocksize *= 2;
        
        _buffersize = _blocksize * _buffernos;
        _lastblock = 0;
        
        StartPlaying = false;
        
        so.playback_rate = Settings.SoundPlaybackRate;
        so.stereo = Settings.Stereo;
        so.sixteen_bit = Settings.SixteenBitSound;
        so.buffer_size = _blocksize;
        so.encoded = FALSE;
        
        if (DirectX_SetSoundMode ())
        {
            SoundBuffer = (uint8 *) malloc (_blocksize * _buffernos + 1024 * 64);
            SoundBuffer += 32 * 1024;
            ZeroMemory (SoundBuffer, _blocksize * _buffernos);
            S9xSetPlaybackRate (so.playback_rate);
        }
        _samplecount = so.sixteen_bit ? _blocksize / 2 : _blocksize;
        _maxsamplecount = SOUND_BUFFER_SIZE;
        if (so.sixteen_bit)
            _maxsamplecount /= 2;
        if (so.stereo)
            _maxsamplecount /= 2;
        
        if (so.samples_mixed_so_far >= _maxsamplecount)
            so.samples_mixed_so_far = 0;
        so.mute_sound = FALSE;

        return (SoundBuffer != NULL);
    }
    else
    {
#endif
        pending_setup = TRUE;
        pending_rate = rate;
        pending_16bit = sixteen_bit;
        pending_stereo = stereo;
        GUI.Stereo = (stereo ? 1 : 0);
		Settings.SoundPlaybackRate = rate;
		Settings.SixteenBitSound = sixteen_bit;
		Settings.Stereo = stereo;
		
		
#ifdef FMOD_SUPPORT
    }
#endif

//	pending_setup = false;
//	return (RealSetupSound (pending_rate, pending_16bit, pending_stereo));
    return (TRUE);
}

bool8 RealSetupSound (long rate, bool8 sixteen_bit, bool8 stereo)
{
    so.mute_sound = TRUE;
    if (SoundBuffer)
    {
        SoundBuffer -= 32 * 1024;
        free ((char *) SoundBuffer);
        SoundBuffer = NULL;
    }

#if 0
    OSVERSIONINFO info;
    info.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
    if (GetVersionEx (&info) && info.dwPlatformId == VER_PLATFORM_WIN32_NT &&
        info.dwMajorVersion == 4)
    {
        _interval = 20;
    }
    else
    {
        _interval = 10;
    }
#endif

    _interval = Settings.SoundMixInterval;

    if (Settings.SoundBufferSize < 1)
        Settings.SoundBufferSize = 1;
    if (Settings.SoundBufferSize > 64)
        Settings.SoundBufferSize = 64;

    _buffernos = 4 * Settings.SoundBufferSize;
    int s = (rate * _interval * 
             (Settings.Stereo ? 2 : 1) * 
             (Settings.SixteenBitSound ? 2 : 1)) / 1000;

    _blocksize = 64;
    while (_blocksize < s)
        _blocksize *= 2;

    _buffersize = _blocksize * _buffernos;
    _lastblock = 0;

    StartPlaying = false;

    so.playback_rate = Settings.SoundPlaybackRate;
    so.stereo = Settings.Stereo;
    so.sixteen_bit = Settings.SixteenBitSound;
    so.buffer_size = _blocksize;
    so.encoded = FALSE;

    if (!DirectX.DSAvailable)
        return (false);

    if (DirectX_SetSoundMode ())
    {
        SoundBuffer = (uint8 *) malloc (_blocksize * _buffernos + 1024 * 64);
        SoundBuffer += 32 * 1024;
        ZeroMemory (SoundBuffer, _blocksize * _buffernos);
        S9xSetPlaybackRate (so.playback_rate);
    }

    _samplecount = so.sixteen_bit ? _blocksize / 2 : _blocksize;
    _maxsamplecount = SOUND_BUFFER_SIZE;
    if (so.sixteen_bit)
        _maxsamplecount /= 2;
    if (so.stereo)
        _maxsamplecount /= 2;

    if (so.samples_mixed_so_far >= _maxsamplecount)
        so.samples_mixed_so_far = 0;
    so.mute_sound = FALSE;

    return (SoundBuffer != NULL);
}


#define FIXED_POINT 0x10000
#define FIXED_POINT_SHIFT 16
#define FIXED_POINT_REMAINDER 0xffff

//static bool8 block_signal = FALSE;
static bool8 block_signal = FALSE;
static volatile bool8 pending_signal = FALSE;

void ProcessSound ();

extern "C" void S9xGenerateSound(void)
{
    if (!SoundBuffer || so.samples_mixed_so_far >= _samplecount ||
        so.mute_sound
#ifdef FMOD_SUPPORT
        || Settings.SoundDriver >= WIN_FMOD_DIRECT_SOUND_DRIVER
#endif
       )
	return;

    block_signal = TRUE;
    so.err_counter += so.err_rate;
    if (so.err_counter >= FIXED_POINT)
    {
        int sample_count = so.err_counter >> FIXED_POINT_SHIFT;
	int byte_offset;
	int byte_count;

        so.err_counter &= FIXED_POINT_REMAINDER;
	if (so.stereo)
	    sample_count <<= 1;
        if (so.samples_mixed_so_far + sample_count > _samplecount)
            sample_count = _samplecount - so.samples_mixed_so_far;

	byte_count = sample_count;

	if (so.sixteen_bit)
	{
	    byte_offset = so.samples_mixed_so_far << 1;
	    byte_count <<= 1;
	}
	else
	    byte_offset = so.samples_mixed_so_far;
	    
        while (sample_count > _maxsamplecount)
        {
            S9xMixSamplesO (SoundBuffer, _maxsamplecount, byte_offset);
            so.samples_mixed_so_far += _maxsamplecount;
            sample_count -= _maxsamplecount;
            byte_offset += SOUND_BUFFER_SIZE;
        }
        S9xMixSamplesO (SoundBuffer, sample_count, byte_offset);
        so.samples_mixed_so_far += sample_count;
    }
    block_signal = FALSE;
    if (pending_signal)
    {
	ProcessSound ();
	pending_signal = FALSE;
    }
}

// Interval ms has passed
void ProcessSound (void)
{
    if (block_signal)
    {
        pending_signal = TRUE;
        return;
    }

    if (pending_setup)
    {
        pending_setup = false;
        RealSetupSound (pending_rate, pending_16bit, pending_stereo);
    }

    if (!DirectX.lpDSB || !SoundBuffer)
        return;

//    bool8 mute = FALSE;

#if 0
    if (DirectX.IdleCount >= GUI.PausedFramesBeforeMutingSound)
    {
        if (!StartPlaying)
        {
            DirectX.lpDSB->Stop ();
            StartPlaying = TRUE;
        }
        return;
    }
#else
//    mute = DirectX.IdleCount >= GUI.PausedFramesBeforeMutingSound;
#endif
    if (StartPlaying)
    {
        if (DirectX.lpDSB->Play (0, 0, DSBPLAY_LOOPING) != DS_OK)
            return;

        _lastblock = 0;
        StartPlaying = false;
    }
    DWORD play_pos = 0, write_pos = 0;
    HRESULT hResult;

    DirectX.lpDSB->GetCurrentPosition (&play_pos, &write_pos);

    DWORD curr_block = ((play_pos / _blocksize) + 1 * Settings.SoundBufferSize) % _buffernos;

//printf ("play_pos = %d, write_pos = %d, curr_block = %d, lastblock = %d\n",
//        play_pos, write_pos, curr_block, _lastblock);
//fflush (stdout);
    if (curr_block != _lastblock)
    {
	BYTE  *B1, *B2;
	DWORD S1, S2;

        write_pos = curr_block * _blocksize;
        _lastblock = curr_block;

        hResult = DirectX.lpDSB->Lock (write_pos, _blocksize, (void **)&B1, 
                                       &S1, (void **)&B2, &S2, 0);
        if (hResult == DSERR_BUFFERLOST)
        {
            DirectX.lpDSB->Restore ();
            hResult = DirectX.lpDSB->Lock (write_pos, _blocksize, 
                                           (void **)&B1, &S1, (void **)&B2, 
                                           &S2, 0);
        }

        if( hResult != DS_OK)
            return;

        if(Settings.Mute || Settings.ForcedPause || 
            Settings.Paused || Settings.StopEmulation)
        {
            if (so.sixteen_bit)
            {
                if (B1)
                    ZeroMemory (B1, S1);
                if (B2)
                    ZeroMemory (B2, S2);
            }
            else
            {
                if (B1)
                    memset (B1, 128, S1);
                if (B2)
                    memset (B2, 128, S2);
            }
        }
        else
        {
            unsigned int sample_count = so.buffer_size;
            int byte_offset;

            if (so.sixteen_bit)
                sample_count >>= 1;

            if (so.samples_mixed_so_far < (int32)sample_count)
            {
                byte_offset = (so.sixteen_bit ? (so.samples_mixed_so_far << 1)
                                              : so.samples_mixed_so_far);

                if (Settings.SoundSync == 2)
                {
                    /*memset (SoundBuffer + (byte_offset & SOUND_BUFFER_SIZE_MASK), 0,
                      sample_count - so.samples_mixed_so_far);*/
                }
                else
                {
                    sample_count -= so.samples_mixed_so_far;

                    while ((long) sample_count > _maxsamplecount)
                    {
                        S9xMixSamplesO (SoundBuffer, _maxsamplecount,
                                        byte_offset);
                        sample_count -= _maxsamplecount;
                        byte_offset += SOUND_BUFFER_SIZE;
                    }
                    S9xMixSamplesO (SoundBuffer, sample_count, byte_offset);
                }
                so.samples_mixed_so_far = 0;
            }
            else
                so.samples_mixed_so_far -= sample_count;

            if (Settings.Mute)
            {
                if (so.sixteen_bit)
                {
                    if (B1)
                        ZeroMemory (B1, S1);
                    if (B2)
                        ZeroMemory (B2, S2);
                }
                else
                {
                    if (B1)
                        memset (B1, 128, S1);
                    if (B2)
                        memset (B2, 128, S2);
                }
            }
            else
            {
                if (B1)
                    memmove (B1, SoundBuffer, S1);
                if (B2)
                    memmove (B2, SoundBuffer + S1, S2);
            }
        }
        hResult = DirectX.lpDSB -> Unlock (B1, S1, B2, S2);
        if( hResult != DS_OK)
            return;
    }
    DWORD Status;
    hResult = DirectX.lpDSB->GetStatus (&Status);
    StartPlaying = !Status;
}

bool8 S9xOpenSoundDevice (int mode, bool8 pStereo, int BufferSize)
{
    return (TRUE);
}

int superscope_pause = 0;
int superscope_turbo = 0;
static RECT dstRect = { 0, 512, 0, 448 };

bool8 S9xReadSuperScopePosition (int &x, int &y, uint32 &buttons)
{
    x = (int) ((GUI.MouseX - dstRect.left) * (256.0 / (dstRect.right - dstRect.left)));
    //not all games are 224
	y = (int) ((GUI.MouseY - dstRect.top) * ((double)PPU.ScreenHeight / (dstRect.bottom - dstRect.top)));
    buttons = GUI.MouseButtons | (superscope_turbo << 2) | 
              (superscope_pause << 3);

	superscope_pause=FALSE;

    return (TRUE);
}

extern unsigned long START;

void S9xSyncSpeed( void)
{
#ifdef NETPLAY_SUPPORT
    if (Settings.NetPlay)
    {
#if defined (NP_DEBUG) && NP_DEBUG == 2
        printf ("CLIENT: SyncSpeed @%d\n", timeGetTime () - START);
#endif
        S9xWinScanJoypads ();
	// Send joypad position update to server
	S9xNPSendJoypadUpdate (joypads [0]);
        LONG prev;
        BOOL success;

	// Wait for heart beat from server
        if ((success = ReleaseSemaphore (GUI.ClientSemaphore, 1, &prev)) &&
            prev == 0)
        {
            // No heartbeats already arrived, have to wait for one.
            // Mop up the ReleaseSemaphore test above...
            WaitForSingleObject (GUI.ClientSemaphore, 0);

            // ... and then wait for the real sync-signal from the
            // client loop thread.
            NetPlay.PendingWait4Sync = WaitForSingleObject (GUI.ClientSemaphore, 100) != WAIT_OBJECT_0;
#if defined (NP_DEBUG) && NP_DEBUG == 2
            if (NetPlay.PendingWait4Sync)
                printf ("CLIENT: PendingWait4Sync1 @%d\n", timeGetTime () - START);
#endif
            IPPU.RenderThisFrame = TRUE;
            IPPU.SkippedFrames = 0;
        }
        else
        {
            if (success)
            {
                // Once for the ReleaseSemaphore above...
                WaitForSingleObject (GUI.ClientSemaphore, 0);
                if (prev == 4 && NetPlay.Waiting4EmulationThread)
                {
                    // Reached the lower behind count threshold - tell the
                    // server its safe to start sending sync pulses again.
                    NetPlay.Waiting4EmulationThread = FALSE;
                    S9xNPSendPause (FALSE);
                }

#if defined (NP_DEBUG) && NP_DEBUG == 2
                if (prev > 1)
                {
                    printf ("CLIENT: SyncSpeed prev: %d @%d\n", prev, timeGetTime () - START);
                }
#endif
            }
            else
            {
#ifdef NP_DEBUG
                printf ("*** CLIENT: SyncSpeed: Release failed @ %d\n", timeGetTime () - START);
#endif
            }

            // ... and again to mop up the already-waiting sync-signal
            NetPlay.PendingWait4Sync = WaitForSingleObject (GUI.ClientSemaphore, 200) != WAIT_OBJECT_0;
#if defined (NP_DEBUG) && NP_DEBUG == 2
            if (NetPlay.PendingWait4Sync)
                printf ("CLIENT: PendingWait4Sync2 @%d\n", timeGetTime () - START);
#endif

	    if (IPPU.SkippedFrames < NetPlay.MaxFrameSkip)
	    {
		IPPU.SkippedFrames++;
		IPPU.RenderThisFrame = FALSE;
	    }
	    else
	    {
		IPPU.RenderThisFrame = TRUE;
		IPPU.SkippedFrames = 0;
	    }
        }
        // Give up remainder of time-slice to any other waiting threads,
        // if they need any time, that is.
        Sleep (0);
        if (!NetPlay.PendingWait4Sync)
        {
            NetPlay.FrameCount++;
            S9xNPStepJoypadHistory ();
        }
    }
    else
#endif

    if ((!Settings.TurboMode && Settings.SkipFrames == AUTO_FRAMERATE)||(XppMovie.Replay_rec == TRUE || XppMovie.Replay_play == TRUE))
    {
        LONG prev;
        BOOL success;
        static unsigned long XPP_FrameTime;

        if ((success = ReleaseSemaphore (GUI.FrameTimerSemaphore, 1, &prev)) &&
            prev == 0)
        {
            // Mop up the ReleaseSemaphore above...
            WaitForSingleObject (GUI.FrameTimerSemaphore, 0);
            // ...and then wait for the real sync signal
            WaitForSingleObject (GUI.FrameTimerSemaphore, 100);
            IPPU.RenderThisFrame = TRUE;
            IPPU.SkippedFrames = 0;
            XPP_FrameTime = 0;
        }
        else
        {
          if (success)
              WaitForSingleObject (GUI.FrameTimerSemaphore, 0);
          WaitForSingleObject (GUI.FrameTimerSemaphore, 0);

	    if (IPPU.SkippedFrames < Settings.AutoMaxSkipFrames && prev > 1)
	    {
		IPPU.SkippedFrames++;
		IPPU.RenderThisFrame = FALSE;
		XPP_FrameTime = 0;
	    }
	    else
	    {

	                if (IPPU.SkippedFrames==0 && XPP_FrameTime > 3)
	                {
		            IPPU.SkippedFrames++;
		            IPPU.RenderThisFrame = FALSE;
		            XPP_FrameTime = 0;
                        }
                        else
                        {
		            IPPU.RenderThisFrame = TRUE;
		            IPPU.SkippedFrames = 0;
		            XPP_FrameTime++;
                        }
	    }
        }
    }
    else
    {
	if (++IPPU.FrameSkip >= (Settings.TurboMode ? Settings.TurboSkipFrames
                                                    : Settings.SkipFrames))
	{
	    IPPU.FrameSkip = 0;
	    IPPU.SkippedFrames = 0;
	    IPPU.RenderThisFrame = TRUE;
	}
	else
	{
	    IPPU.SkippedFrames++;
	    IPPU.RenderThisFrame = FALSE;
	}
    }
}

const char *S9xBasename (const char *f)
{
	const char *p;
	const char *pp = f;
	size_t slen = lstrlen(f);
	for (p = f; p < (f + slen); p++) {
		if(*(p + 1) && ((*(unsigned char *)p >= 0x81 && *(unsigned char *)p <= 0x9f) || (*(unsigned char *)p >= 0xe0 && *(unsigned char *)p <= 0xfc)) &&
			((*(unsigned char *)(p + 1) >= 0x40 && *(unsigned char *)(p + 1) <= 0x7e) || (*(unsigned char *)(p + 1) >= 0x80 && *(unsigned char *)(p + 1) <= 0xfc)))
			++p;
		else if (*p == (const char)'\\')
			pp = (p + 1);
	}
	return pp;
/*
    const char *p;
    if ((p = strrchr (f, '/')) != NULL || (p = strrchr (f, '\\')) != NULL)
	return (p + 1);

#ifdef __DJGPP
    if (p = strrchr (f, SLASH_CHAR))
	return (p + 1);
#endif

    return (f);
*/
}

bool8 S9xReadMousePosition (int which, int &x, int &y, uint32 &buttons)
{
    if (which == 0)
    {
        x = GUI.MouseX;
        y = GUI.MouseY;
        buttons = GUI.MouseButtons;
        return (TRUE);
    }

    return (FALSE);
}

PALETTEENTRY S9xPaletteEntry[256];
void S9xSetPalette( void)
{
    LPDIRECTDRAWPALETTE lpDDTemp;
    uint16 Brightness = IPPU.MaxBrightness * 140;

    // Only update the palette structures if needed
    if( GUI.ScreenDepth == 8)
    {
        if (Settings.SixteenBit)
        {
            for (int i = 0; i < 256; i++)
            {
                S9xPaletteEntry[i].peRed   = FixedColours [i].red;
                S9xPaletteEntry[i].peGreen = FixedColours [i].green;
                S9xPaletteEntry[i].peBlue  = FixedColours [i].blue;
            }
        }
        else
        {
            for (int i = 0; i < 256; i ++)
            {	
                S9xPaletteEntry[i].peRed   = (((PPU.CGDATA [i] >>  0) & 0x1F) * Brightness) >> 8;
                S9xPaletteEntry[i].peGreen = (((PPU.CGDATA [i] >>  5) & 0x1F) * Brightness) >> 8;
                S9xPaletteEntry[i].peBlue  = (((PPU.CGDATA [i] >> 10) & 0x1F) * Brightness) >> 8;
            }
        }
		if( DirectX.lpDDSPrimary != NULL)
		{
			DirectX.lpDDSPrimary -> GetPalette( &lpDDTemp);
			if( lpDDTemp != DirectX.lpDDPalette)
				DirectX.lpDDSPrimary -> SetPalette( DirectX.lpDDPalette);
		}
		else
		{
			DirectX.lpDDSPrimary2 -> GetPalette( &lpDDTemp);
			if( lpDDTemp != DirectX.lpDDPalette)
				DirectX.lpDDSPrimary2 -> SetPalette( DirectX.lpDDPalette);
		}
        DirectX.lpDDPalette->SetEntries (0, 0, 256, S9xPaletteEntry);
    }
}

bool LockSurface (LPDIRECTDRAWSURFACE lpDDSurface, SSurface *lpSurface)
{
    DDSURFACEDESC ddsd;
    HRESULT hResult;
    int retry;

    ddsd.dwSize = sizeof( ddsd);

    retry = 0;
    while( true)
    {
        hResult = lpDDSurface -> Lock( NULL, &ddsd, DDLOCK_WAIT, NULL);

        if( hResult == DD_OK)
        {
            lpSurface -> Width = ddsd.dwWidth;
            lpSurface -> Height = ddsd.dwHeight;
            lpSurface -> Pitch = ddsd.lPitch;
            lpSurface -> Surface = (unsigned char *)ddsd.lpSurface;
            return true;
        }

        if( hResult == DDERR_SURFACELOST)
        {
            retry++;
            if( retry > 5)
            {	return false;	}

            hResult = lpDDSurface -> Restore();
            GUI.ScreenCleared = true;
            if (hResult != DD_OK)
            {	return false;	}

            continue;
        }

        if( hResult != DDERR_WASSTILLDRAWING)
            return false;
    }
}

bool LockSurface2 (LPDIRECTDRAWSURFACE2 lpDDSurface, SSurface *lpSurface)
{
    DDSURFACEDESC ddsd;
    HRESULT hResult;
    int retry;

    ddsd.dwSize = sizeof( ddsd);

    retry = 0;
    while( true)
    {
        hResult = lpDDSurface -> Lock( NULL, &ddsd, DDLOCK_WAIT, NULL);

        if( hResult == DD_OK)
        {
            lpSurface -> Width = ddsd.dwWidth;
            lpSurface -> Height = ddsd.dwHeight;
            lpSurface -> Pitch = ddsd.lPitch;
            lpSurface -> Surface = (unsigned char *)ddsd.lpSurface;
            return true;
        }

        if( hResult == DDERR_SURFACELOST)
        {
            retry++;
            if( retry > 5)
            {	return false;	}

            hResult = lpDDSurface -> Restore();
            GUI.ScreenCleared = true;
            if (hResult != DD_OK)
            {	return false;	}

            continue;
        }

        if( hResult != DDERR_WASSTILLDRAWING)
            return false;
    }
}

BYTE ScreenBuffer[512*478*2 + (512*3*2)];//2xSaI engine access 512*3*2
BYTE SubScreenBuffer[512*478*2];
BYTE ZBuffer[512*478];
BYTE SubZBuffer[512*478];
//static bool8 locked_surface = FALSE;
bool8 locked_surface = FALSE;
static SSurface Dst;

bool8 S9xInitUpdate( void)
{
    GFX.SubScreen = SubScreenBuffer;
    GFX.ZBuffer = ZBuffer;
    GFX.SubZBuffer = SubZBuffer;

    if (GUI.Scale <= 1 && 
//        (GUI.Stretch || !GUI.FullScreen) &&
        !GUI.NeedDepthConvert &&
//        !DirectX.DoubleBuffered &&
        ((DirectX.lpDDSOffScreen2 && 
          LockSurface2 (DirectX.lpDDSOffScreen2, &Dst))))
    {
        GFX.RealPitch = GFX.Pitch2 = GFX.Pitch = Dst.Pitch;
        GFX.Screen = Dst.Surface;
        DirectX.lpDDSOffScreen2->Unlock (Dst.Surface);
        locked_surface = TRUE;

		GFX.ZPitch = GFX.Pitch;
		if (Settings.SixteenBit)
			GFX.ZPitch >>= 1;
    }
    else
    {
        locked_surface = FALSE;
//        GFX.RealPitch = GFX.Pitch = 512 * 2;
//        GFX.Screen = &ScreenBuffer[512*3*2];
		if (Settings.SixteenBit)
		{
			GFX.RealPitch = GFX.Pitch2 = GFX.Pitch = 512 * 2;
			GFX.ZPitch = GFX.Pitch;
			GFX.ZPitch >>= 1;
			GFX.Screen = &ScreenBuffer[512*3*2];
		}
		else
		{
			GFX.RealPitch = GFX.Pitch2 = GFX.Pitch = 512;
			GFX.ZPitch = GFX.Pitch;
			GFX.Screen = &ScreenBuffer[512*3*2];
			GFX.SubScreen = &ScreenBuffer[(512*3*2) + (512*478)];
		}
    }
    GFX.Delta = (GFX.SubScreen - GFX.Screen) >> 1;

    if (Settings.SixteenBit)
    {
        GFX.PPL = GFX.Pitch >> 1;
        GFX.PPLx2 = GFX.Pitch;
    }
    else
    {
        GFX.PPL = GFX.Pitch;
        GFX.PPLx2 = GFX.Pitch * 2;
    }
//    GFX.Pitchx2 = GFX.Pitch * 2;
    return (TRUE);
}

extern int CurFullscreenDisplayModeIndex;
extern unsigned char dRendering;
extern "C" unsigned char HiresModeDetectedFlag;
BYTE DepthConvertBuf[512 * 478 * 2];
static SSurface DepthConvertSurface = {NULL, 512 * 2, 512, 478};

bool8 S9xDeinitUpdate( int Width, int Height, bool8 sixteen_bit)
{
    SSurface Src;
    LPDIRECTDRAWSURFACE2 lpDDSurface2 = NULL;
    bool PrimarySurfaceLockFailed = false;
    RECT srcRect;

#if 0
HRESULT Render3DEnvironment ();

Render3DEnvironment ();
return true;
#endif

	if(Options.DisplayScreenBufferAll && GUI.Scale) {
		Width = 512;
		Height = 478;
	}
    Src.Width = Width;
    Src.Height = Height;
    Src.Pitch = GFX.Pitch;
    Src.Surface = &ScreenBuffer[512*3*2];

	ExtendHeight = Src.Height == 239 || Src.Height == 478;
	DoubleHeight = Src.Height > 240;
	DoubleWidth  = Src.Width == 512;

	//Directly Render to screen
	if((!GUI.Scale || (GUI.Scale > 1 && GUI.Scale < 4)) && dRendering && !DirectX.lpDDClipper && !GUI.Stretch) {
		Src.Surface = GFX.Screen;
		//Clear screen
		if(ExtendHeight_bak && !ExtendHeight)
		{
			if(DirectX.Windowed)
			{
				RECT clrRect;
				HDC clrhdc;
				GetClientRect (GUI.hWnd, &clrRect);
				clrhdc = GetDC(GUI.hWnd);
				FillRect(clrhdc, &clrRect, (HBRUSH)GetStockObject(BLACK_BRUSH));
				ReleaseDC(GUI.hWnd, clrhdc);
			}
			else
			{
				DirectX_ClearPrimarySurface();
				DirectX_ClearBackbufferSurface();
			}
		}

		if(GUI.hWnd != GetForegroundWindow() ||
			(!DirectX.Windowed && ExtendHeight &&
			(CurFullscreenDisplayModeIndex <= 1  ||
			(GUI.Scale && (CurFullscreenDisplayModeIndex == 12 || CurFullscreenDisplayModeIndex == 13))))) {
			ExtendHeight_bak = ExtendHeight;
			return TRUE;
		}

		srcRect.top    = 0;
		srcRect.bottom = Height;
		srcRect.left   = 0;
		srcRect.right  = Width;

		SelectRenderMethodd ();

		GetClientRect (GUI.hWnd, &dstRect);
		int width = Width;
		int height = Height;
#if 0
		if (GUI.Scale == 1)
		{
			width = 512;
			if (height < 240)
				height *= 2;
		}
#endif
		 if (GUI.Scale > 0)
		{
			if (width < 257)
				width *= 2;
			if (height < 240)
				height *= 2;
		}
		POINT p;
		p.x = ((dstRect.right - dstRect.left) - width) / 2;
		p.y = ((dstRect.bottom - dstRect.top) - height) / 2;
		ClientToScreen (GUI.hWnd, &p);

		dstRect.top = p.y;
		dstRect.left = p.x;
		dstRect.bottom = dstRect.top + height;
		dstRect.right  = dstRect.left + width;

		if(!DirectX.Windowed && DirectX.DoubleBuffered) {
			if(DirectX.lpDDSBack2 && LockSurface2 (DirectX.lpDDSBack2, &Dst));
			else if(DirectX.lpDDSBack && LockSurface (DirectX.lpDDSBack, &Dst));
			else {
				ExtendHeight_bak = ExtendHeight;
				return false;
			}
			RenderMethodd (&Src, &Dst, &srcRect, &dstRect);
			if(DirectX.lpDDSBack2) {
				DirectX.lpDDSBack2->Unlock( Dst.Surface);
				DirectX.lpDDSPrimary2 -> Flip( NULL, DDFLIP_WAIT);
			}
			else {
				DirectX.lpDDSBack->Unlock( Dst.Surface);
				DirectX.lpDDSPrimary -> Flip( NULL, DDFLIP_WAIT);
			}
		}
		else {
			if(DirectX.lpDDSPrimary2 && LockSurface2 (DirectX.lpDDSPrimary2, &Dst));
			else if(DirectX.lpDDSPrimary && LockSurface (DirectX.lpDDSPrimary, &Dst));
			else {
				ExtendHeight_bak = ExtendHeight;
				return false;
			}
			if(DirectX.wait_for_vsync)
				DirectX.lpDD->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN,0);
			RenderMethodd (&Src, &Dst, &srcRect, &dstRect);
			if(DirectX.lpDDSPrimary2)
				DirectX.lpDDSPrimary2->Unlock( Dst.Surface);
			else
				DirectX.lpDDSPrimary->Unlock( Dst.Surface);
		}
		ExtendHeight_bak = ExtendHeight;
		return true;
	}

    if (locked_surface)
    {
        lpDDSurface2 = DirectX.lpDDSOffScreen2;
        PrimarySurfaceLockFailed = true;
    }
    else
    {
        if (DirectX.lpDDSBack2 == NULL)
            lpDDSurface2 = DirectX.lpDDSPrimary2;
        else 
            lpDDSurface2 = DirectX.lpDDSBack2;
    }

    if (!locked_surface
#if 0
     && !VOODOO_MODE && !OPENGL_MODE
#endif
     )
    {
//        if (GUI.Stretch || GUI.Scale == 1 || !GUI.FullScreen || DirectX.DoubleBuffered || !LockSurface2 (lpDDSurface2, &Dst))
//        if (GUI.Stretch || GUI.Scale > 0 || DirectX.Windowed || DirectX.DoubleBuffered || !LockSurface2 (lpDDSurface2, &Dst))
//ItXN[\tKvǂAStretch̏ꍇBlt\tg傪KvAnC]c{ {iHHjgc
        if (GUI.Stretch || DirectX.Windowed || GUI.Scale > 0 || !LockSurface2 (lpDDSurface2, &Dst))
        {
            lpDDSurface2 = DirectX.lpDDSOffScreen2;
            if (!LockSurface2 (lpDDSurface2, &Dst))
                return false;
            
            PrimarySurfaceLockFailed = true;
        }
    }

	if (!locked_surface)
	{
		if (GUI.NeedDepthConvert)
		{
			RECT drect;
			if (GUI.Scale <= 1)
			{
				srcRect.left = srcRect.top = 0;
				srcRect.right = Width;
				srcRect.bottom = Height;

				drect.left = (Dst.Width - Src.Width) >> 1;
				drect.right = drect.left + Src.Width;
				drect.top = (Dst.Height - Src.Height) >> 1;
				drect.bottom = drect.top + Src.Height;

				ConvertDepth (&Src, &Dst, &srcRect, &drect);
				srcRect = drect;
			}
			//SoftWareX2 Scanline
			else if(GUI.Scale <= 3)
			{
				if(!DoubleWidth && !DoubleHeight)
				{
					SelectRenderMethodd ();
					srcRect.top    = 0;
					srcRect.bottom = Height;
					srcRect.left   = 0;
					srcRect.right  = Width;
					drect.left = (Dst.Width - (Width << 1)) >> 1;
					drect.right = drect.left + (Width << 1);
					drect.top = (Dst.Height - (Height << 1)) >> 1;
					drect.bottom = drect.top + (Height << 1);
					RenderMethodd (&Src, &Dst, &srcRect, &drect);
					srcRect = drect;
				}
				else
				{
					//nC][hȍ
					HiresModeDetectedFlag = TRUE;
					//nC]\tgEGAg傷邩H
					if(DirectX.DisableHardwareScale) {
						SelectRenderMethodd ();
						srcRect.top    = 0;
						srcRect.bottom = Height;
						srcRect.left   = 0;
						srcRect.right  = Width;
						int width = ((Width < 257) ? (Width << 1) : Width);
						int height = ((Height < 240) ? (Height << 1) : Height);
						drect.left = (Dst.Width - width) >> 1;
						drect.right = drect.left + width;
						drect.top = (Dst.Height - height) >> 1;
						drect.bottom = drect.top + height;
						RenderMethodd (&Src, &Dst, &srcRect, &drect);
						srcRect = drect;
					}
					else {
						srcRect.left = srcRect.top = 0;
						srcRect.right = Width;
						srcRect.bottom = Height;

						drect.left = (Dst.Width - Src.Width) >> 1;
						drect.right = drect.left + Src.Width;
						drect.top = (Dst.Height - Src.Height) >> 1;
						drect.bottom = drect.top + Src.Height;

						ConvertDepth (&Src, &Dst, &srcRect, &drect);
						srcRect = drect;
					}
				}
			}
			else
			{
				if(!DoubleWidth && !DoubleHeight)
				{
					DepthConvertSurface.Surface = DepthConvertBuf;
					SelectRenderMethod ();
					RenderMethod (Src, DepthConvertSurface, &srcRect);

					drect.left = (Dst.Width - (srcRect.right - srcRect.left)) >> 1;
					drect.right = drect.left + (srcRect.right - srcRect.left);
					drect.top = (Dst.Height - (srcRect.bottom - srcRect.top)) >> 1;
					drect.bottom = drect.top + (srcRect.bottom - srcRect.top);

					ConvertDepth (&DepthConvertSurface, &Dst, &srcRect, &drect);

					srcRect = drect;
				}
				else
				{
					//nC][hȍ
					HiresModeDetectedFlag = TRUE;
					//nC]\tgEGAg傷邩H
					if(DirectX.DisableHardwareScale) {
						SelectRenderMethodd ();
						srcRect.top    = 0;
						srcRect.bottom = Height;
						srcRect.left   = 0;
						srcRect.right  = Width;
						int width = ((Width < 257) ? (Width << 1) : Width);
						int height = ((Height < 240) ? (Height << 1) : Height);
						drect.left = (Dst.Width - width) >> 1;
						drect.right = drect.left + width;
						drect.top = (Dst.Height - height) >> 1;
						drect.bottom = drect.top + height;
						RenderMethodd (&Src, &Dst, &srcRect, &drect);
						srcRect = drect;
					}
					else {
						srcRect.left = srcRect.top = 0;
						srcRect.right = Width;
						srcRect.bottom = Height;

						drect.left = (Dst.Width - Src.Width) >> 1;
						drect.right = drect.left + Src.Width;
						drect.top = (Dst.Height - Src.Height) >> 1;
						drect.bottom = drect.top + Src.Height;

						ConvertDepth (&Src, &Dst, &srcRect, &drect);
						srcRect = drect;
					}
				}
			}
		}
		else {
			//nC]\tgEGAg傷邩H
			if((DoubleWidth || DoubleHeight) && DirectX.DisableHardwareScale)
			{
				RECT drect;
				HiresModeDetectedFlag = TRUE;
				SelectRenderMethodd ();
				srcRect.top    = 0;
				srcRect.bottom = Height;
				srcRect.left   = 0;
				srcRect.right  = Width;
				int width = ((Width < 257) ? (Width << 1) : Width);
				int height = ((Height < 240) ? (Height << 1) : Height);
				drect.left = (Dst.Width - width) >> 1;
				drect.right = drect.left + width;
				drect.top = (Dst.Height - height) >> 1;
				drect.bottom = drect.top + height;
				RenderMethodd (&Src, &Dst, &srcRect, &drect);
				srcRect = drect;
			}
			else {
				SelectRenderMethod ();
				RenderMethod (Src, Dst, &srcRect);
			}
		}
    }
    else
    {
        srcRect.top    = 0;
        srcRect.bottom = Height;
        srcRect.left   = 0;
        srcRect.right  = Width;
    }
#if 0
    if (!VOODOO_MODE && !OPENGL_MODE)
    {
#endif
        if( PrimarySurfaceLockFailed)
        {
            POINT p;

            if (GUI.Stretch)
            {
                GetClientRect (GUI.hWnd, &dstRect);
                p.x = p.y = 0;
                ClientToScreen (GUI.hWnd, &p);
                dstRect.right = (dstRect.right - dstRect.left) + p.x;
                dstRect.bottom = (dstRect.bottom - dstRect.top) + p.y;
                dstRect.top = p.y;
                dstRect.left = p.x;
            }
            else
            {
                GetClientRect (GUI.hWnd, &dstRect);
                int width = srcRect.right - srcRect.left;
                int height = srcRect.bottom - srcRect.top;

/*
                if (GUI.Scale == 1)
                {
                    width = 512;
                    if (height < 240)
                        height *= 2;
                }
*/
                if (GUI.Scale > 0)
                {
                    if (width < 257)
                        width *= 2;
                    if (height < 240)
                        height *= 2;
                }
                p.x = ((dstRect.right - dstRect.left) - width) / 2;
                p.y = ((dstRect.bottom - dstRect.top) - height) / 2;
                ClientToScreen (GUI.hWnd, &p);

                dstRect.top = p.y;
                dstRect.left = p.x;
                dstRect.bottom = dstRect.top + height;
                dstRect.right  = dstRect.left + width;
            }
        }
		else {
			dstRect = srcRect;
		}
        lpDDSurface2->Unlock( Dst.Surface);	

		//Clear screen
		if(ExtendHeight_bak && !ExtendHeight)
		{
			if(DirectX.Windowed)
			{
				RECT clrRect;
				HDC clrhdc;
				GetClientRect (GUI.hWnd, &clrRect);
				clrhdc = GetDC(GUI.hWnd);
				FillRect(clrhdc, &clrRect, (HBRUSH)GetStockObject(BLACK_BRUSH));
				ReleaseDC(GUI.hWnd, clrhdc);
			}
			else
			{
				DirectX_ClearPrimarySurface();
				DirectX_ClearBackbufferSurface();
			}
		}

		ExtendHeight_bak = ExtendHeight;

		if(GUI.hWnd != GetForegroundWindow() ||
			(!DirectX.Windowed && ExtendHeight &&
			(CurFullscreenDisplayModeIndex <= 1  ||
			(GUI.Scale && (CurFullscreenDisplayModeIndex == 12 || CurFullscreenDisplayModeIndex == 13)))))
			return TRUE;

        if (PrimarySurfaceLockFailed)
        {
            if( DirectX.lpDDSBack2 == NULL)
                lpDDSurface2 = DirectX.lpDDSPrimary2;
            else 
                lpDDSurface2 = DirectX.lpDDSBack2;
            if(DirectX.wait_for_vsync && !DirectX.DoubleBuffered)
            	DirectX.lpDD->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN,0);
            if (lpDDSurface2->Blt (&dstRect, DirectX.lpDDSOffScreen2, &srcRect, DDBLT_WAIT, NULL) != DD_OK)
                return false;
        }
        if(!DirectX.Windowed && DirectX.DoubleBuffered)
        	DirectX.lpDDSPrimary2 -> Flip( NULL, DDFLIP_WAIT);
#if 0
    }
    else
        dstRect = srcRect;
#endif

    return true;
}

bool InitSnes9X( void)
{
#ifdef DEBUGGER
    extern FILE *trace;

    trace = fopen( "SNES9X.TRC", "wt");
    freopen( "SNES9X.OUT", "wt", stdout);
    freopen( "SNES9X.ERR", "wt", stderr);

//    CPU.Flags |= TRACE_FLAG;
//    APU.Flags |= TRACE_FLAG;
#endif

#ifdef GENERATE_OFFSETS_H
    offsets_h = fopen ("offsets.h", "wt");
    generate_offsets_h (0, NULL);
    fclose (offsets_h);
#endif

    if(!Memory.Init())
		return FALSE;

    GFX.Pitch = 512 * 2;
    GFX.RealPitch = 512 * 2;
    GFX.Screen = &ScreenBuffer[512*3*2];
    GFX.SubScreen = SubScreenBuffer;
    GFX.Delta = (GFX.SubScreen - GFX.Screen) >> 1;

    S9xSetWinPixelFormat ();

    if(!S9xGraphicsInit())
    	return FALSE;

	S9xInitUpdate();

    if(!S9xInitAPU())
    	return FALSE;

    S9xInitSound (7, true, _blocksize);

    so.playback_rate = Settings.SoundPlaybackRate;
    so.stereo = Settings.Stereo;
    so.sixteen_bit = Settings.SixteenBitSound;
    so.buffer_size = _blocksize;


    // Sound options
    Settings.SoundBufferSize = Settings.SoundBufferSize;

#ifdef USE_GLIDE
    Settings.GlideEnable = FALSE;
#endif

	return TRUE;
}

int ffs (uint32 mask)
{
    int m = 0;
    if (mask)
    {
        while (!(mask & (1 << m)))
            m++;

        return (m);
    }

    return (0);
}

void S9xSetWinPixelFormat ()
{
    extern int Init_2xSaI (uint32 BitFormat);

    S9xSetRenderPixelFormat (RGB565);
    Init_2xSaI (565);
    GUI.NeedDepthConvert = FALSE;

#if 0
    if (VOODOO_MODE)
    {
        GUI.ScreenDepth = 16;
        GUI.RedShift = 0;
        GUI.GreenShift = 5;
        GUI.BlueShift = 11;
        Settings.SixteenBit = TRUE;
    }
    else
    if (OPENGL_MODE)
    {
        GUI.ScreenDepth = 16;
        GUI.RedShift = 0;
        GUI.GreenShift = 5;
        GUI.BlueShift = 11;
        Settings.SixteenBit = TRUE;
    }
    else
    {
#endif
        GUI.ScreenDepth = DirectX.DDPixelFormat.dwRGBBitCount;
        if (GUI.ScreenDepth == 15)
            GUI.ScreenDepth = 16;

        GUI.RedShift = ffs (DirectX.DDPixelFormat.dwRBitMask);
        GUI.GreenShift = ffs (DirectX.DDPixelFormat.dwGBitMask);
        GUI.BlueShift = ffs (DirectX.DDPixelFormat.dwBBitMask);

        if((DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
           GUI.ScreenDepth == 16 &&
           DirectX.DDPixelFormat.dwRBitMask == 0xF800 &&
           DirectX.DDPixelFormat.dwGBitMask == 0x07E0 &&
           DirectX.DDPixelFormat.dwBBitMask == 0x001F)
        {
            S9xSetRenderPixelFormat (RGB565);
            Init_2xSaI (565);
        }
        else
            if( (DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
                GUI.ScreenDepth == 16 &&
                DirectX.DDPixelFormat.dwRBitMask == 0x7C00 &&
                DirectX.DDPixelFormat.dwGBitMask == 0x03E0 &&
                DirectX.DDPixelFormat.dwBBitMask == 0x001F)
            {
                S9xSetRenderPixelFormat (RGB555);
                Init_2xSaI (555);
            }
            else
                if((DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
                   GUI.ScreenDepth == 16 &&
                   DirectX.DDPixelFormat.dwRBitMask == 0x001F &&
                   DirectX.DDPixelFormat.dwGBitMask == 0x07E0 &&
                   DirectX.DDPixelFormat.dwBBitMask == 0xF800)
                {
                    S9xSetRenderPixelFormat (BGR565);
                    Init_2xSaI (565);
                }
                else
                    if( (DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
                        GUI.ScreenDepth == 16 &&
                        DirectX.DDPixelFormat.dwRBitMask == 0x001F &&
                        DirectX.DDPixelFormat.dwGBitMask == 0x03E0 &&
                        DirectX.DDPixelFormat.dwBBitMask == 0x7C00)
                    {
                        S9xSetRenderPixelFormat (BGR555);
                        Init_2xSaI (555);
                    }
                    else
                        if (DirectX.DDPixelFormat.dwRGBBitCount == 8 ||
                            DirectX.DDPixelFormat.dwRGBBitCount == 24 ||
                            DirectX.DDPixelFormat.dwRGBBitCount == 32)
                        {
                            S9xSetRenderPixelFormat (RGB565);
                            Init_2xSaI (565);
                        }
        
        if (
#if 0
            !VOODOO_MODE &&
            !OPENGL_MODE &&
            (
#endif
             (GUI.ScreenDepth == 8 && Settings.SixteenBit) ||
             (GUI.ScreenDepth == 16 && !Settings.SixteenBit) ||
             (GUI.ScreenDepth == 24 || GUI.ScreenDepth == 32))
            GUI.NeedDepthConvert = TRUE;
        
        if (Settings.SixteenBit && 
            (GUI.ScreenDepth == 24 || GUI.ScreenDepth == 32))
        {
            GUI.RedShift += 3;
            GUI.GreenShift += 3;
            GUI.BlueShift += 3;
        }
#if 0
    }
#endif

    int l = 0;
    int i;

    for (i = 0; i < 6; i++)
    {
	int r = (i * 31) / (6 - 1);
	for (int j = 0; j < 6; j++)
	{
	    int g = (j * 31) / (6 - 1);
	    for (int k = 0; k < 6; k++)
	    { 
		int b = (k * 31) / (6 - 1);

		FixedColours [l].red = r << 3;
		FixedColours [l].green = g << 3;
		FixedColours [l++].blue = b << 3;
	    }
	}
    }

    //int *color_diff = new int [0x10000];
    int diffr, diffg, diffb, maxdiff = 0, won = 0, lost;
    int r, d = 8;
    for (r = 0; r <= (int) MAX_RED; r++)
    {
	int cr, g, q;
      
	int k = 6 - 1;
	cr = (r * k) / MAX_RED;
	q  = (r * k) % MAX_RED;
	if (q > d && cr < k) 
	    cr++;
	diffr = abs (cr * k - r);
	for (g = 0; g <= (int) MAX_GREEN; g++)
	{
	    int cg, b;
	  
	    k  = 6 - 1;
	    cg = (g * k) / MAX_GREEN;
	    q  = (g * k) % MAX_GREEN;
	    if(q > d && cg < k)
		cg++;
	    diffg = abs (cg * k - g);
	    for (b = 0; b <= (int) MAX_BLUE; b++) 
	    {
		int cb;
		int rgb = BUILD_PIXEL2(r, g, b);

		k  = 6 - 1;
		cb = (b * k) / MAX_BLUE;
		q  = (b * k) % MAX_BLUE;
		if (q > d && cb < k)
		    cb++;
		diffb = abs (cb * k - b);
		S9xwPalette[rgb] = (cr * 6 + cg) * 6 + cb;
		color_diff[rgb] = diffr + diffg + diffb;
		if (color_diff[rgb] > maxdiff)
		    maxdiff = color_diff[rgb];
	    }
	}
    }

    while (maxdiff > 0 && l < 256)
    {
	int newmaxdiff = 0;
	lost = 0; won++;
	for (r = MAX_RED; r >= 0; r--)
	{
	    int g;
      
	    for (g = MAX_GREEN; g >= 0; g--)
	    {
		int b;
	  
		for (b = MAX_BLUE; b >= 0; b--) 
		{
		    int rgb = BUILD_PIXEL2(r, g, b);

		    if (color_diff[rgb] == maxdiff)
		    {
			if (l >= 256)
			    lost++;
			else
			{
			    FixedColours [l].red = r << 3;
			    FixedColours [l].green = g << 3;
			    FixedColours [l].blue = b << 3;
			    S9xwPalette [rgb] = l++;
			}
			color_diff[rgb] = 0;
		    }
		    else
			if (color_diff[rgb] > newmaxdiff)
			    newmaxdiff = color_diff[rgb];
		    
		}
	    }
	}
	maxdiff = newmaxdiff;
    }
    free (color_diff);
}

void S9xAutoSaveSRAM ()
{
    Memory.SaveSRAM (S9xGetFilename (".srm"));
}

void S9xSetPause (uint32 mask)
{
    Settings.ForcedPause |= mask;
}

void S9xClearPause (uint32 mask)
{
    Settings.ForcedPause &= ~mask;
    if (!Settings.ForcedPause)
    {
        // Wake up the main loop thread just if its blocked in a GetMessage
        // call.
        PostMessage (GUI.hWnd, WM_NULL, 0, 0);
    }
}

void DeinitSnes9X( void)
{
#ifdef DEBUGGER
    extern FILE *trace;
	if(trace)
		fclose(trace);
#endif
	if(trace_fs)
		fclose(trace_fs);
	if (SoundBuffer)
	{
		SoundBuffer -= 32 * 1024;
		free ((char *) SoundBuffer);
		SoundBuffer = NULL;
	}
	Memory.Deinit();
	S9xGraphicsDeinit();
	S9xDeinitAPU();

	//j[폜
	if(!GUI.AppMenuConnect) {
		DestroyMenu(GUI.hMenu);
	}
	if(GUI.hRecentGamesMenu)
		DestroyMenu(GUI.hRecentGamesMenu);
	if(GUI.hVolMenu)
		DestroyMenu(GUI.hVolMenu);

	DirectInput_exit();
	//destroy DirectX
	DirectX_DeInitialize();
}

static int S9xCompareSDD1IndexEntries (const void *p1, const void *p2)
{
    return (*(uint32 *) p1 - *(uint32 *) p2);
}

void S9xLoadSDD1Data ()
{
    char filename [_MAX_PATH + 1];
    char index [_MAX_PATH + 1];
    char data [_MAX_PATH + 1];

//	Settings.SDD1Pack=FALSE;
    Memory.FreeSDD1Data ();

//    strcpy (filename, GUI.FreezeFileDir);

//	if(strlen(GUI.FreezeFileDir)!=0)
     strcpy (filename, GUI.FreezeFileDir);
/*
	else
	{
	    char dir [_MAX_DIR + 1];
	    char drive [_MAX_DRIVE + 1];
		char name [_MAX_FNAME + 1];
	    char ext [_MAX_EXT + 1];
		_splitpath(Memory.ROMFilename,drive, dir, name, ext);
		_makepath(filename,drive, dir, NULL, NULL);
	}
*/

 if(!Settings.SDD1Pack)
 {
    if (strncmp (Memory.ROMName, "Star Ocean", 10) == 0)
	strcat (filename, "\\socnsdd1");
    else
    if (strncmp (Memory.ROMName, "STREET FIGHTER ZERO2", 20) == 0)
	strcat (filename, "\\sfz2sdd1"); // (JAPAN)
    else
    if (strncmp (Memory.ROMName, "STREET FIGHTER ALPHA2", 21) == 0 && Memory.ROM [0x7fd9] == 1)
	strcat (filename, "\\sfa2sdd1"); // (USA)
    else
    if (strncmp (Memory.ROMName, "STREET FIGHTER ALPHA2", 21) == 0 && Memory.ROM [0x7fd9] == 2)
	strcat (filename, "\\sf2esdd1"); // (PAL)
    else
	strcat (filename, "\\unknown");

    strcpy (index, filename);
    strcat (index, "\\SDD1GFX.IDX");
    strcpy (data, filename);
    strcat (data, "\\SDD1GFX.DAT");

    FILE *fs = fopen (index, "rb");
    int len = 0;
    
    if (fs)
    {
        // Index is stored as a sequence of entries, each entry being
        // 12 bytes consisting of:
        // 4 byte key: (24bit address & 0xfffff * 16) | translated block
        // 4 byte ROM offset
        // 4 byte length
        fseek (fs, 0, SEEK_END);
        len = ftell (fs);
        rewind (fs);
        Memory.SDD1Index = (uint8 *) malloc (len);
        fread (Memory.SDD1Index, 1, len, fs);
        fclose (fs);
        Memory.SDD1Entries = len / 12;
        
        if (!(fs = fopen (data, "rb")))
        {
            free ((char *) Memory.SDD1Index);
            Memory.SDD1Index = NULL;
            Memory.SDD1Entries = 0;
        }
        else
        {
            fseek (fs, 0, SEEK_END);
            len = ftell (fs);
            rewind (fs);
            Memory.SDD1Data = (uint8 *) malloc (len);
            fread (Memory.SDD1Data, 1, len, fs);
            fclose (fs);
            
            qsort (Memory.SDD1Index, Memory.SDD1Entries, 12,
                   S9xCompareSDD1IndexEntries);
        }
    }
 }

}

extern "C"
{
	char*osd_GetPackDir()
	{
		static char filename[MAX_PATH];
		memset(filename, 0, MAX_PATH);
		
		if(strlen(GUI.FreezeFileDir)!=0)
			strcpy (filename, GUI.FreezeFileDir);
		else
		{
			char dir [_MAX_DIR + 1];
			char drive [_MAX_DRIVE + 1];
			char name [_MAX_FNAME + 1];
			char ext [_MAX_EXT + 1];
			_splitpath(Memory.ROMFilename,drive, dir, name, ext);
			_makepath(filename,drive, dir, NULL, NULL);
		}
		
		if(!strncmp((char*)&Memory.ROM [0xffc0], "SUPER POWER LEAG 4   ", 21))
		{
			strcat(filename, "\\SPL4-SP7");
		}
		else if(!strncmp((char*)&Memory.ROM [0xffc0], "MOMOTETSU HAPPY      ",21))
		{
			strcat(filename, "\\SMHT-SP7");
		}
		else if(!strncmp((char*)&Memory.ROM [0xffc0], "HU TENGAI MAKYO ZERO ", 21))
		{
			strcat(filename, "\\FEOEZSP7");
		}
		else if(!strncmp((char*)&Memory.ROM [0xffc0], "JUMP TENGAIMAKYO ZERO",21))
		{
			strcat(filename, "\\SJUMPSP7");
		}
		else strcat(filename, "\\MISC-SP7");
		return filename;
	}
}

bool JustifierOffscreen()
{
	return (bool)((GUI.MouseButtons&2)!=0);
}

void JustifierButtons(uint32& justifiers)
{
	{
		if((GUI.MouseButtons&1)||(GUI.MouseButtons&2))
		{
			justifiers|=0x00100;
		}
		if(GUI.MouseButtons&4)
		{
			justifiers|=0x00400;
		}
	}
}

